# Copyright (C) 2008 John Paulett (john -at- paulett.org)
# Copyright (C) 2009-2024 David Aguilar (davvid -at- gmail.com)
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

import collections
import os
import warnings

import pytest

import jsonpickle
import jsonpickle.backend
import jsonpickle.handlers
from jsonpickle import tags, util


class ListLike:
    def __init__(self):
        self.internal_list = []

    def append(self, item):
        self.internal_list.append(item)

    def __getitem__(self, i):
        return self.internal_list[i]


class Thing:
    def __init__(self, name):
        self.name = name
        self.child = None

    def __iter__(self):
        for attr in [
            x for x in getattr(self.__class__, '__dict__') if not x.startswith('__')
        ]:
            yield attr, getattr(self, attr)

    def __repr__(self):
        return 'Thing("%s")' % self.name


class Capture:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __repr__(self):
        return object.__repr__(self) + (f'({self.args!r}, {self.kwargs!r})')


class ThingWithProps:
    def _get_identity(self):
        keys = [self.dogs, self.monkies, self.name]
        return hash('-'.join([str(key) for key in keys]))

    identity = property(_get_identity)

    def _get_dogs(self):
        return self._critters[0][1]

    dogs = property(_get_dogs)

    def _get_monkies(self):
        return self._critters[1][1]

    monkies = property(_get_monkies)

    def __init__(self, name='', dogs='reliable', monkies='tricksy'):
        self.name = name
        self._critters = (('dogs', dogs), ('monkies', monkies))

    def __eq__(self, other):
        return self.identity == other.identity

    def __getstate__(self):
        out = dict(
            __identity__=self.identity,
            nom=self.name,
            dogs=self.dogs,
            monkies=self.monkies,
        )
        return out

    def __setstate__(self, state_dict):
        self._critters = (
            ('dogs', state_dict.get('dogs')),
            ('monkies', state_dict.get('monkies')),
        )
        self.name = state_dict.get('nom', '')
        ident = state_dict.get('__identity__')
        if ident != self.identity:
            raise ValueError('expanded object does not match original state!')


class UserDict(dict):
    """A user class that inherits from :class:`dict`"""

    def __init__(self, **kwargs):
        dict.__init__(self, **kwargs)
        self.valid = False


class Outer:
    class Middle:
        class Inner:
            pass


class MySlots:
    __slots__ = ('alpha', '__beta')

    def __init__(self):
        self.alpha = 1
        self.__beta = 1


class MyPropertiesSlots:
    __slots__ = ('alpha', 'arr')

    def __init__(self):
        self.alpha = 1
        self.arr = [1, 2, 3]

    @property
    def arr_len(self):
        return len(self.arr)

    @property
    def other_object(self):
        return Thing('Test')

    def __eq__(self, other):
        return (
            self.alpha == other.alpha
            and self.arr == other.arr
            and self.arr_len == other.arr_len
            and list(self.other_object) == list(other.other_object)
        )


class MyPropertiesDict:
    def __init__(self):
        self.alpha = 1
        self.arr = [1, 2, 3]

    @property
    def arr_len(self):
        return len(self.arr)

    @property
    def other_object(self):
        return Thing('Test')

    def __eq__(self, other):
        return (
            self.alpha == other.alpha
            and self.arr == other.arr
            and self.arr_len == other.arr_len
            and list(self.other_object) == list(other.other_object)
        )


# see issue #478
class SafeData:
    __slots__ = ()


class SafeString(str, SafeData):
    __slots__ = ()


def on_missing_callback(class_name):
    # not actually a runtime problem but it doesn't matter
    warnings.warn('The unpickler could not find %s' % class_name, RuntimeWarning)


@pytest.fixture
def pickler():
    """Returns a default-constructed pickler"""
    return jsonpickle.pickler.Pickler()


@pytest.fixture
def unpickler():
    """Returns a default-constructed unpickler"""
    return jsonpickle.unpickler.Unpickler()


@pytest.fixture
def b85_pickler():
    """Returns a pickler setup for base85 encoding"""
    return jsonpickle.pickler.Pickler(use_base85=True)


def test_bytes_base85(b85_pickler):
    """base85 is emitted when the pickler is setup to do so"""
    data = os.urandom(16)
    encoded = util.b85encode(data)
    assert b85_pickler.flatten(data) == {tags.B85: encoded}


def test_bytes_base64_default(pickler):
    """base64 must be used by default"""
    data = os.urandom(16)
    encoded = util.b64encode(data)
    assert pickler.flatten(data) == {tags.B64: encoded}


def test_decode_base85(unpickler):
    """base85 data must be restored"""
    expected = 'Pÿthöñ 3!'.encode()
    pickled = {tags.B85: util.b85encode(expected)}
    assert unpickler.restore(pickled) == expected


@pytest.mark.parametrize('value', ['', '/', '\udc00', 1, True, False, None, [], {}])
def test_decode_invalid_b85(value, unpickler):
    """Invalid base85 data restores to an empty string"""
    expected = b''
    pickled = {tags.B85: value}
    assert unpickler.restore(pickled) == expected


def test_base85_still_handles_base64(unpickler):
    """base64 must be restored even though base85 is the default"""
    expected = 'Pÿthöñ 3!'.encode()
    pickled = {tags.B64: util.b64encode(expected)}
    assert unpickler.restore(pickled) == expected


@pytest.mark.parametrize(
    'value', ['', 'x', '!', '\udc00', 0, 1, True, False, None, [], {}]
)
def test_decode_invalid_b64(value, unpickler):
    """Invalid base85 data restores to an empty string"""
    expected = b''
    pickled = {tags.B64: value}
    assert unpickler.restore(pickled) == expected


def test_string(pickler, unpickler):
    """Strings must roundtrip"""
    assert pickler.flatten('a string') == 'a string'
    assert unpickler.restore('a string') == 'a string'


def test_int(pickler, unpickler):
    """Ints must roundtrip"""
    assert pickler.flatten(3) == 3
    assert unpickler.restore(3) == 3


def test_float(pickler, unpickler):
    """Floats must roundtrip"""
    assert pickler.flatten(3.5) == 3.5
    assert unpickler.restore(3.5) == 3.5


def test_boolean(pickler, unpickler):
    """Booleans must roundtrip"""
    assert pickler.flatten(True) is True
    assert pickler.flatten(False) is False
    assert unpickler.restore(True) is True
    assert unpickler.restore(False) is False


def test_none(pickler, unpickler):
    """None must roundtrip"""
    assert pickler.flatten(None) is None
    assert unpickler.restore(None) is None


def test_list(pickler, unpickler):
    """Nested and mixed lists must roundtrip"""
    # multiple types of values
    list_a = [1, 35.0, 'value']
    assert pickler.flatten(list_a) == list_a
    assert unpickler.restore(list_a) == list_a
    # nested list
    list_b = [40, 40, list_a, 6]
    assert pickler.flatten(list_b) == list_b
    assert unpickler.restore(list_b) == list_b
    # 2D list
    list_c = [[1, 2], [3, 4]]
    assert pickler.flatten(list_c) == list_c
    assert unpickler.restore(list_c) == list_c
    # empty list
    list_d = []
    assert pickler.flatten(list_d) == list_d
    assert unpickler.restore(list_d) == list_d


def test_nonetype():
    """NoneType must roundtrip"""
    typ = type(None)
    typ_pickled = jsonpickle.encode(typ)
    typ_unpickled = jsonpickle.decode(typ_pickled)
    assert typ_unpickled == typ


def test_set(pickler, unpickler):
    """Validate the internal representation for set() objects"""
    set_list = ['orange', 'apple', 'grape']
    set_obj = set(set_list)
    flattened = pickler.flatten(set_obj)
    for s in set_list:
        assert s in flattened[tags.SET]
    set_pickle = {tags.SET: set_list}
    assert unpickler.restore(set_pickle) == set_obj


@pytest.mark.parametrize('value', ['', 0, 1, True, False, None, [], {}])
def test_set_with_invalid_data(value, unpickler):
    """Invalid serialized set data results in an empty set"""
    data = {tags.SET: value}
    result = unpickler.restore(data)
    assert result == set()


@pytest.mark.parametrize('value', ['', 0, 1, True, False, None, [], {}])
def test_tuple_with_invalid_data(value, unpickler):
    """Invalid serialized tuple data results in an empty tuple"""
    data = {tags.TUPLE: value}
    result = unpickler.restore(data)
    assert result == tuple()


def test_iterator_with_invalid_data(unpickler):
    """Invalid serialized iterator data results in an empty iterator"""
    data = {tags.ITERATOR: set()}
    result = unpickler.restore(data)
    with pytest.raises(StopIteration):
        next(result)
        assert False


@pytest.mark.parametrize(
    'value', ['', 0, True, False, None, [], {}, ('x',), {'x': True}]
)
def test_reduce_with_invalid_data(value, unpickler):
    """Invalid serialized reduce data results in an empty list"""
    data = {tags.REDUCE: value}
    result = unpickler.restore(data)
    assert result == []


@pytest.mark.parametrize('value', ['', 'x', 1, True, [], {}])
def test_restore_id_with_invalid_data(value, unpickler):
    """Invalid serialized ID data results in None"""
    result = unpickler.restore({'ref': {tags.ID: value}})
    assert result['ref'] is None


def test_dict(pickler, unpickler):
    """Our custom keys are preserved when user dicts contain them"""
    dict_a = {'key1': 1.0, 'key2': 20, 'key3': 'thirty', tags.JSON_KEY + '6': 6}
    assert pickler.flatten(dict_a) == dict_a
    assert unpickler.restore(dict_a) == dict_a
    dict_b = {}
    assert pickler.flatten(dict_b) == dict_b
    assert unpickler.restore(dict_b) == dict_b


def test_tuple(pickler, unpickler):
    """Validate the internal represntation for tuples"""
    # currently all collections are converted to lists
    tuple_a = (4, 16, 32)
    tuple_a_pickle = {tags.TUPLE: [4, 16, 32]}
    assert pickler.flatten(tuple_a) == tuple_a_pickle
    assert unpickler.restore(tuple_a_pickle) == tuple_a
    tuple_b = (4,)
    tuple_b_pickle = {tags.TUPLE: [4]}
    assert pickler.flatten(tuple_b) == tuple_b_pickle
    assert unpickler.restore(tuple_b_pickle) == tuple_b


def _roundtrip(data):
    """Encode and decode and object through jsonpickle"""
    return jsonpickle.decode(jsonpickle.encode(data))


def test_tuple_roundtrip():
    """Tuples can roundtrip"""
    data = (1, 2, 3)
    assert _roundtrip(data) == data


def test_set_roundtrip():
    """Sets can roundtrip"""
    data = {1, 2, 3}
    assert _roundtrip(data) == data


def test_list_roundtrip():
    """Lists can roundtrip"""
    data = [1, 2, 3]
    assert _roundtrip(data) == data


def test_class(pickler, unpickler):
    """Nested objects can roundtrip"""
    inst = Thing('test name')
    inst.child = Thing('child name')
    flattened = pickler.flatten(inst)
    assert flattened['name'] == 'test name'
    child = flattened['child']
    assert child['name'] == 'child name'
    inflated = unpickler.restore(flattened)
    assert inflated.name == 'test name'
    assert type(inflated) is Thing
    assert inflated.child.name == 'child name'
    assert type(inflated.child) is Thing


def test_classlist(pickler, unpickler):
    """Lists with mixed object values can roundtrip"""
    array = [Thing('one'), Thing('two'), 'a string']
    flattened = pickler.flatten(array)
    assert 'one' == flattened[0]['name']
    assert 'two' == flattened[1]['name']
    assert 'a string' == flattened[2]
    inflated = unpickler.restore(flattened)
    assert 'one' == inflated[0].name
    assert type(inflated[0]) is Thing
    assert 'two' == inflated[1].name
    assert type(inflated[1]) is Thing
    assert 'a string' == inflated[2]


def test_classdict(pickler, unpickler):
    """Dicts with object values can roundtrip"""
    obj_dict = {'k1': Thing('one'), 'k2': Thing('two'), 'k3': 3}
    flattened = pickler.flatten(obj_dict)
    assert 'one' == flattened['k1']['name']
    assert 'two' == flattened['k2']['name']
    assert 3 == flattened['k3']
    inflated = unpickler.restore(flattened)
    assert 'one' == inflated['k1'].name
    assert type(inflated['k1']) is Thing
    assert 'two' == inflated['k2'].name
    assert type(inflated['k2']) is Thing
    assert 3 == inflated['k3']


def test_recursive():
    """create a recursive structure and test that we can handle it"""
    parent = Thing('parent')
    child = Thing('child')
    child.sibling = Thing('sibling')
    parent.self = parent
    parent.child = child
    parent.child.twin = child
    parent.child.parent = parent
    parent.child.sibling.parent = parent
    cloned = jsonpickle.decode(jsonpickle.encode(parent))
    assert parent.name == cloned.name
    assert parent.child.name == cloned.child.name
    assert parent.child.sibling.name == cloned.child.sibling.name
    assert cloned == cloned.child.parent
    assert cloned == cloned.child.sibling.parent
    assert cloned == cloned.child.twin.parent
    assert cloned.child == cloned.child.twin


def test_tuple_notunpicklable(pickler):
    """Tuples become lists when unpicklable is False"""
    pickler.unpicklable = False
    flattened = pickler.flatten(('one', 2, 3))
    assert flattened == ['one', 2, 3]


def test_set_not_unpicklable(pickler):
    """Sets become lists when unpicklable is False"""
    pickler.unpicklable = False
    flattened = pickler.flatten({'one', 2, 3})
    assert 'one' in flattened
    assert 2 in flattened
    assert 3 in flattened
    assert isinstance(flattened, list)


def test_thing_with_module(pickler, unpickler):
    """Objects with references to modules can roundtrip"""
    obj = Thing('with-module')
    obj.themodule = os
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert inflated.themodule == os


def test_thing_with_module_safe(pickler, unpickler):
    """Objects with inner references to modules can roundtrip in safe mode"""
    obj = Thing('with-module')
    obj.themodule = os
    flattened = pickler.flatten(obj)
    unpickler.safe = True
    inflated = unpickler.restore(flattened)
    assert inflated.themodule is os
    # Unsafe mode
    unpickler.safe = False
    inflated = unpickler.restore(flattened)
    assert inflated.themodule is os


def test_thing_with_submodule(pickler, unpickler):
    """Objects with inner references to modules can roundtrip"""
    obj = Thing('with-submodule')
    obj.submodule = collections
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert inflated.submodule == collections


def test_type_reference(pickler, unpickler):
    """This test ensures that users can store references to types."""
    obj = Thing('object-with-type-reference')
    # reference the built-in 'object' type
    obj.typeref = object
    flattened = pickler.flatten(obj)
    assert flattened['typeref'] == {tags.TYPE: 'builtins.object'}
    inflated = unpickler.restore(flattened)
    assert inflated.typeref == object


def test_class_reference(pickler, unpickler):
    """This test ensures that users can store references to classes."""
    obj = Thing('object-with-class-reference')
    # reference the 'Thing' class (not an instance of the class)
    obj.classref = Thing
    flattened = pickler.flatten(obj)
    assert flattened['classref'] == {tags.TYPE: 'jsonpickle_test.Thing'}
    inflated = unpickler.restore(flattened)
    assert inflated.classref is Thing


def test_supports_getstate_setstate(pickler, unpickler):
    """Objects with getstate and setstate can roundtrip"""
    obj = ThingWithProps('object-which-defines-getstate-setstate')
    flattened = pickler.flatten(obj)
    assert flattened[tags.STATE].get('__identity__')
    assert flattened[tags.STATE].get('nom')
    inflated = unpickler.restore(flattened)
    assert obj == inflated


def test_references(pickler, unpickler):
    """References in lists can roundtrip"""
    obj_a = Thing('foo')
    obj_b = Thing('bar')
    coll = [obj_a, obj_b, obj_b]
    flattened = pickler.flatten(coll)
    inflated = unpickler.restore(flattened)
    assert len(inflated) == len(coll)
    for x in range(len(coll)):
        assert repr(coll[x]) == repr(inflated[x])


def test_references_in_number_keyed_dict(pickler, unpickler):
    """Dicts with numbers as keys and objects as values can roundtrip

    Because JSON must coerce integers to strings in dict keys, the sort
    order may have a tendency to change between pickling and unpickling,
    and this could affect the object references.
    """
    one = Thing('one')
    two = Thing('two')
    twelve = Thing('twelve')
    two.child = twelve
    obj = {
        1: one,
        2: two,
        12: twelve,
    }
    assert list(sorted(obj.keys())) != list(map(int, sorted(map(str, obj.keys()))))
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert len(inflated) == 3
    assert inflated['12'].name == 'twelve'


def test_builtin_error():
    """Builtin errors can roundtrip"""
    expect = AssertionError
    json = jsonpickle.encode(expect)
    actual = jsonpickle.decode(json)
    assert expect == actual
    assert expect is actual


def test_builtin_function():
    """Builtin functions can roundtrip"""
    expect = dir
    json = jsonpickle.encode(expect)
    actual = jsonpickle.decode(json)
    assert expect == actual
    assert expect is actual


def test_restore_legacy_builtins():
    """Decoding is backwards compatible.

    jsonpickle 0.9.6 and earlier used the Python 2 `__builtin__`
    naming for builtins. Ensure those can be loaded until they're
    no longer supported.
    """
    ae = jsonpickle.decode('{"py/type": "__builtin__.AssertionError"}')
    assert ae is AssertionError
    ae = jsonpickle.decode('{"py/type": "exceptions.AssertionError"}')
    assert ae is AssertionError
    cls = jsonpickle.decode('{"py/type": "__builtin__.int"}')
    assert cls is int


@pytest.mark.parametrize(
    'value,expect',
    [
        ('module_does_not_exist/ignored', None),
        ('builtins/int', None),
        ('builtins/invalid.int', None),
        ('builtins/builtinsx.int', None),
    ],
)
def test_restore_invalid_repr(value, expect, unpickler):
    """Test restoring invalid repr tags"""
    result = unpickler.restore({tags.REPR: value})
    assert result is expect


def test_unpickler_on_missing():
    """Emit warnings when decoding objects whose classes are missing"""
    encoded = jsonpickle.encode(Outer.Middle.Inner())
    assert isinstance(jsonpickle.decode(encoded), Outer.Middle.Inner)
    # Alter the encoded string to create cases where the class is missing
    # at multiple levels.
    assert encoded == '{"py/object": "jsonpickle_test.Outer.Middle.Inner"}'
    missing_cases = [
        '{"py/object": "MISSING.Outer.Middle.Inner"}',
        '{"py/object": "jsonpickle_test.MISSING.Middle.Inner"}',
        '{"py/object": "jsonpickle_test.Outer.MISSING.Inner"}',
        '{"py/object": "jsonpickle_test.Outer.Middle.MISSING"}',
    ]
    for case in missing_cases:
        # https://docs.python.org/3/library/warnings.html#testing-warnings
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")
            jsonpickle.decode(case, on_missing='warn')
            assert issubclass(w[-1].category, UserWarning)
            assert 'Unpickler._restore_object could not find' in str(w[-1].message)
            jsonpickle.decode(case, on_missing=on_missing_callback)
            assert issubclass(w[-1].category, RuntimeWarning)
            assert 'The unpickler could not find' in str(w[-1].message)

        assert jsonpickle.decode(
            case, on_missing='ignore'
        ) == jsonpickle.backend.json.loads(case)
        try:
            jsonpickle.decode(case, on_missing='error')
        except jsonpickle.errors.ClassNotFoundError:
            # it's supposed to error
            assert True
        else:
            assert False


def test_private_slot_members():
    """Objects with private slot members can roundtrip"""
    obj = jsonpickle.loads(jsonpickle.dumps(MySlots()))
    alpha = getattr(obj, 'alpha', '(missing alpha)')
    beta = getattr(obj, '_' + obj.__class__.__name__ + '__beta', '(missing beta)')
    assert alpha == beta


def test_include_properties_slots():
    """Properties of slot-using objects are encoded with include_properties=True"""
    obj = MyPropertiesSlots()
    dumped = jsonpickle.dumps(obj, include_properties=True)
    assert 'py/property' in dumped
    assert jsonpickle.loads(dumped) == obj


def test_include_properties_dict():
    """Dicts with properties are encoded with include_properties=True"""
    obj = MyPropertiesDict()
    dumped = jsonpickle.dumps(obj, include_properties=True)
    assert 'py/property' in dumped
    assert jsonpickle.loads(dumped) == obj


def test_load_non_fully_qualified_classes():
    """Custom classes can be specified using string names when decoding"""
    # reuse MyPropertiesSlots because it has a nice eq method
    obj = MyPropertiesSlots()
    encoded = jsonpickle.encode(obj)
    # MyPropertiesSlots and MyPropertiesDict have compatible eq methods
    decoded = jsonpickle.decode(
        encoded, classes={'MyPropertiesSlots': MyPropertiesDict}
    )
    assert isinstance(decoded, MyPropertiesDict)
    assert obj == decoded


def test_classes_dict():
    """Custom classes can be used when decoding"""
    # reuse MyPropertiesSlots because it has a nice eq method
    obj = MyPropertiesSlots()
    encoded = jsonpickle.encode(obj)
    # MyPropertiesSlots and MyPropertiesDict have compatible eq methods
    decoded = jsonpickle.decode(encoded, classes={MyPropertiesSlots: MyPropertiesDict})
    assert isinstance(decoded, MyPropertiesDict)
    assert obj == decoded


def test_warnings():
    """Emit warnings when pickling file descriptors"""
    data = os.fdopen(os.pipe()[0])
    with warnings.catch_warnings(record=True) as w:
        jsonpickle.encode(data, warn=True)
        assert len(w) == 1
        assert issubclass(w[-1].category, UserWarning)
        assert "replaced with None" in str(w[-1].message)


def test_API_names():
    """Enforce expected names in main module"""
    names = list(vars(jsonpickle))
    assert 'pickler' in names
    assert 'unpickler' in names
    assert 'JSONBackend' in names
    assert '__version__' in names
    assert 'register' in names
    assert 'unregister' in names
    assert 'Pickler' in names
    assert 'Unpickler' in names
    assert 'encode' in names
    assert 'decode' in names


def test_encode():
    """Encode an object with default settings"""
    expect = Thing('A name')
    pickle = jsonpickle.encode(expect)
    actual = jsonpickle.decode(pickle)
    assert expect.name == actual.name
    assert expect.child == actual.child


def test_encode_notunpicklable():
    """Encode an object with unpicklable=False"""
    obj = Thing('A name')
    expect = {'name': 'A name', 'child': None}
    pickle = jsonpickle.encode(obj, unpicklable=False)
    actual = jsonpickle.decode(pickle)
    assert expect['name'] == actual['name']


def test_decode():
    """Validate that we can decode a known json payload"""
    expect = Thing('A name')
    expected_json = (
        '{"%s": "jsonpickle_test.Thing", "name": "A name", "child": null}' % tags.OBJECT
    )
    actual = jsonpickle.decode(expected_json)
    assert expect.name == actual.name
    assert type(expect) is type(actual)


def test_json():
    """Validate the json representation"""
    expect = Thing('A name')
    expected_json = (
        '{"%s": "jsonpickle_test.Thing", "name": "A name", "child": null}' % tags.OBJECT
    )
    pickle = jsonpickle.encode(expect)
    actual = jsonpickle.decode(pickle)
    assert actual.name == expect.name
    assert actual.child == expect.child
    actual = jsonpickle.decode(expected_json)
    assert expect.name == actual.name
    assert type(expect) is type(actual)


def test_unicode_dict_keys():
    """Non-ascii unicode data in dict keys is preserved"""
    uni = chr(0x1234)
    pickle = jsonpickle.encode({uni: uni})
    actual = jsonpickle.decode(pickle)
    assert uni in actual
    assert actual[uni] == uni


def test_tuple_dict_keys_default():
    """Test that we handle dictionaries with tuples as keys."""
    tuple_dict = {(1, 2): 3, (4, 5): {(7, 8): 9}}
    pickle = jsonpickle.encode(tuple_dict)
    expect = {'(1, 2)': 3, '(4, 5)': {'(7, 8)': 9}}
    actual = jsonpickle.decode(pickle)
    assert expect == actual
    tuple_dict = {(1, 2): [1, 2]}
    pickle = jsonpickle.encode(tuple_dict)
    actual = jsonpickle.decode(pickle)
    assert actual['(1, 2)'] == [1, 2]


def test_tuple_dict_keys_with_keys_enabled():
    """Test that we handle dictionaries with tuples as keys."""
    tuple_dict = {(1, 2): 3, (4, 5): {(7, 8): 9}}
    pickle = jsonpickle.encode(tuple_dict, keys=True)
    expect = tuple_dict
    actual = jsonpickle.decode(pickle, keys=True)
    assert expect == actual
    tuple_dict = {(1, 2): [1, 2]}
    pickle = jsonpickle.encode(tuple_dict, keys=True)
    actual = jsonpickle.decode(pickle, keys=True)
    assert actual[(1, 2)] == [1, 2]


def test_None_dict_key_default():
    """None is stringified by default when used as a dict key"""
    expect = {'null': None}
    obj = {None: None}
    pickle = jsonpickle.encode(obj)
    actual = jsonpickle.decode(pickle)
    assert expect == actual


def test_None_dict_key_with_keys_enabled():
    """None can be used as a dict key when keys=True"""
    expect = {None: None}
    obj = {None: None}
    pickle = jsonpickle.encode(obj, keys=True)
    actual = jsonpickle.decode(pickle, keys=True)
    assert expect == actual


def test_object_dict_keys():
    """Test that we handle random objects as keys."""
    thing = Thing('random')
    pickle = jsonpickle.encode({thing: True})
    actual = jsonpickle.decode(pickle)
    assert actual == {'Thing("random")': True}


def test_int_dict_keys_defaults():
    """Int keys are stringified by default"""
    int_dict = {1000: [1, 2]}
    pickle = jsonpickle.encode(int_dict)
    actual = jsonpickle.decode(pickle)
    assert actual['1000'] == [1, 2]


def test_int_dict_keys_with_keys_enabled():
    """Int keys are supported with keys=True"""
    int_dict = {1000: [1, 2]}
    pickle = jsonpickle.encode(int_dict, keys=True)
    actual = jsonpickle.decode(pickle, keys=True)
    assert actual[1000] == [1, 2]


def test_string_key_requiring_escape_dict_keys_with_keys_enabled():
    """Dict keys that require escaping are handled"""
    json_key_dict = {tags.JSON_KEY + '6': [1, 2]}
    pickled = jsonpickle.encode(json_key_dict, keys=True)
    unpickled = jsonpickle.decode(pickled, keys=True)
    assert unpickled[tags.JSON_KEY + '6'] == [1, 2]


def test_string_key_not_requiring_escape_dict_keys_with_keys_enabled():
    """String keys that do not require escaping are not escaped"""
    str_dict = {'name': [1, 2]}
    pickled = jsonpickle.encode(str_dict, keys=True)
    unpickled = jsonpickle.decode(pickled)
    assert 'name' in unpickled


def test_dict_subclass():
    """Dict subclasses can roundtrip"""
    obj = UserDict()
    obj.valid = True
    obj.s = 'string'
    obj.d = 'd_string'
    obj['d'] = {}
    obj['s'] = 'test'
    pickle = jsonpickle.encode(obj)
    actual = jsonpickle.decode(pickle)
    assert type(actual) is UserDict
    assert 'd' in actual
    assert 's' in actual
    assert hasattr(actual, 'd')
    assert hasattr(actual, 's')
    assert hasattr(actual, 'valid')
    assert obj['d'] == actual['d']
    assert obj['s'] == actual['s']
    assert obj.d == actual.d
    assert obj.s == actual.s
    assert obj.valid == actual.valid


def test_dict_subclass_with_references():
    """Dict subclasses with references can roundtrip"""
    d = UserDict()
    d.s = 'string'
    d['s'] = 'test'
    obj = [d, d, d.__dict__]
    pickle = jsonpickle.encode(obj)
    actual = jsonpickle.decode(pickle)
    assert type(actual) is list
    assert type(actual[0]) is UserDict
    assert actual[0].s == 'string'
    assert actual[0]['s'] == 'test'
    assert actual[0] == actual[1]
    assert actual[0].__dict__ == actual[2]


def test_list_of_objects():
    """Using the same object in a list is preserved and not duplicated"""
    a = Thing('a')
    b = Thing('b')
    pickle = jsonpickle.encode([a, b, b])
    actual = jsonpickle.decode(pickle)
    assert actual[1] == actual[2]
    assert type(actual[0]) is Thing
    assert actual[0].name == 'a'
    assert actual[1].name == 'b'
    assert actual[2].name == 'b'


def test_refs_keys_values():
    """Objects in dict keys can roundtrip"""
    j = Thing('random')
    object_dict = {j: j}
    pickle = jsonpickle.encode(object_dict, keys=True)
    actual = jsonpickle.decode(pickle, keys=True)
    assert list(actual.values()) == list(actual.keys())


def test_object_keys_to_list():
    """Objects in dict values have their references preserved"""
    j = Thing('random')
    object_dict = {j: [j, j]}
    pickle = jsonpickle.encode(object_dict, keys=True)
    actual = jsonpickle.decode(pickle, keys=True)
    obj = list(actual.keys())[0]
    assert obj.name == j.name
    assert obj is actual[obj][0]
    assert obj is actual[obj][1]


def test_refs_in_objects():
    """Objects in lists have their references preserved"""
    a = Thing('a')
    b = Thing('b')
    pickle = jsonpickle.encode([a, b, b])
    actual = jsonpickle.decode(pickle)
    assert actual[0] != actual[1]
    assert actual[1] == actual[2]
    assert actual[1] is actual[2]


def test_refs_recursive():
    """Complicated recursive refs can roundtrip"""
    a = Thing('a')
    a.self_list = [Thing('0'), Thing('1'), Thing('2')]
    a.first = a.self_list[0]
    a.stuff = {a.first: a.first}
    a.morestuff = {a.self_list[1]: a.stuff}
    pickle = jsonpickle.encode(a, keys=True)
    b = jsonpickle.decode(pickle, keys=True)
    item = b.self_list[0]
    assert item == b.first
    assert item == b.stuff[b.first]
    assert item == b.morestuff[b.self_list[1]][b.first]


def test_load_backend():
    """Ensures that load_backend() is working"""
    assert jsonpickle.load_backend('simplejson', 'dumps', 'loads', ValueError)


def test_set_preferred_backend_allows_magic():
    """Pluggable backends can be used"""
    backend = 'os.path'
    jsonpickle.load_backend(backend, 'split', 'join', AttributeError)
    jsonpickle.set_preferred_backend(backend)
    slash_hello, world = jsonpickle.encode('/hello/world')
    jsonpickle.remove_backend(backend)

    assert slash_hello == '/hello'
    assert world == 'world'


def test_load_backend_submodule():
    """Inner modules can be used as a backend"""
    jsonpickle.load_backend('os.path', 'split', 'join', AttributeError)
    assert (
        'os.path' in jsonpickle.json._backend_names
        and 'os.path' in jsonpickle.json._encoders
        and 'os.path' in jsonpickle.json._decoders
        and 'os.path' in jsonpickle.json._encoder_options
        and 'os.path' in jsonpickle.json._decoder_exceptions
    )
    jsonpickle.remove_backend('os.path')


def _backend_is_partially_loaded(backend):
    """Return True if the specified backend is incomplete"""
    return (
        backend in jsonpickle.json._backend_names
        or backend in jsonpickle.json._encoders
        or backend in jsonpickle.json._decoders
        or backend in jsonpickle.json._encoder_options
        or backend in jsonpickle.json._decoder_exceptions
    )


def test_load_backend_handles_bad_encode():
    """Ignore bad encoders"""
    load_backend = jsonpickle.load_backend
    assert not load_backend('os.path', 'bad!', 'split', AttributeError)
    assert not _backend_is_partially_loaded('os.path')


def test_load_backend_raises_on_bad_decode():
    """Ignore bad decoders"""
    load_backend = jsonpickle.load_backend
    assert not load_backend('os.path', 'join', 'bad!', AttributeError)
    assert not _backend_is_partially_loaded('os.path')


def test_load_backend_handles_bad_loads_exc():
    """Ignore bad decoder exceptions"""
    load_backend = jsonpickle.load_backend
    assert not load_backend('os.path', 'join', 'split', 'bad!')
    assert not _backend_is_partially_loaded('os.path')


def test_list_item_reference():
    """References to objects inside lists are preserved"""
    thing = Thing('parent')
    thing.child = Thing('child')
    thing.child.refs = [thing]
    encoded = jsonpickle.encode(thing)
    decoded = jsonpickle.decode(encoded)
    assert id(decoded) == id(decoded.child.refs[0])


def test_reference_to_list():
    """References to lists are preserved"""
    thing = Thing('parent')
    thing.a = [1]
    thing.b = thing.a
    thing.b.append(thing.a)
    thing.b.append([thing.a])
    encoded = jsonpickle.encode(thing)
    decoded = jsonpickle.decode(encoded)
    assert 1 == decoded.a[0]
    assert 1 == decoded.b[0]
    assert id(decoded.b) == id(decoded.a)
    assert id(decoded.a[1]) == id(decoded.a)
    assert id(decoded.a[2][0]) == id(decoded.a)


def test_make_refs_disabled_list():
    """Lists are duplicated when make_refs is False"""
    obj_a = Thing('foo')
    obj_b = Thing('bar')
    coll = [obj_a, obj_b, obj_b]
    encoded = jsonpickle.encode(coll, make_refs=False)
    decoded = jsonpickle.decode(encoded)

    assert len(decoded) == 3
    assert decoded[0] is not decoded[1]
    assert decoded[1] is not decoded[2]


def test_make_refs_disabled_reference_to_list():
    """References are lost when maek_refs is False"""
    thing = Thing('parent')
    thing.a = [1]
    thing.b = thing.a
    thing.b.append(thing.a)
    thing.b.append([thing.a])
    encoded = jsonpickle.encode(thing, make_refs=False)
    decoded = jsonpickle.decode(encoded)
    assert decoded.a[0] == 1
    assert decoded.b[0] == 1
    assert decoded.a[1][0:3] == '[1,'  # ]
    assert decoded.a[2][0][0:3] == '[1,'  # ]


def test_can_serialize_inner_classes():
    """Specify hidden classes when decoding"""

    class InnerScope:
        """Private class visible to this method only"""

        def __init__(self, name):
            self.name = name

    obj = InnerScope('test')
    encoded = jsonpickle.encode(obj)
    # Single class
    decoded = jsonpickle.decode(encoded, classes=InnerScope)
    _test_inner_class(InnerScope, obj, decoded)
    # List of classes
    decoded = jsonpickle.decode(encoded, classes=[InnerScope])
    _test_inner_class(InnerScope, obj, decoded)
    # Tuple of classes
    decoded = jsonpickle.decode(encoded, classes=(InnerScope,))
    _test_inner_class(InnerScope, obj, decoded)
    # Set of classes
    decoded = jsonpickle.decode(encoded, classes={InnerScope})
    _test_inner_class(InnerScope, obj, decoded)


def _test_inner_class(cls, obj, decoded):
    assert isinstance(obj, cls)
    assert decoded.name == obj.name


def test_can_serialize_nested_classes():
    """Instances of nested classes can roundtrip"""
    middle = Outer.Middle
    inner = Outer.Middle.Inner
    encoded_middle = jsonpickle.encode(middle)
    encoded_inner = jsonpickle.encode(inner)
    decoded_middle = jsonpickle.decode(encoded_middle)
    decoded_inner = jsonpickle.decode(encoded_inner)

    assert isinstance(decoded_middle, type)
    assert isinstance(decoded_inner, type)
    assert decoded_middle == middle
    assert decoded_inner == inner


def test_can_serialize_nested_class_objects():
    """Class objects in an inner scope can roundtrip"""
    middle_obj = Outer.Middle()
    middle_obj.attribute = 5
    inner_obj = Outer.Middle.Inner()
    inner_obj.attribute = 6
    encoded_middle_obj = jsonpickle.encode(middle_obj)
    encoded_inner_obj = jsonpickle.encode(inner_obj)
    decoded_middle_obj = jsonpickle.decode(encoded_middle_obj)
    decoded_inner_obj = jsonpickle.decode(encoded_inner_obj)
    assert isinstance(decoded_middle_obj, Outer.Middle)
    assert isinstance(decoded_inner_obj, Outer.Middle.Inner)
    assert middle_obj.attribute == decoded_middle_obj.attribute
    assert inner_obj.attribute == decoded_inner_obj.attribute


def test_listlike():
    """List-like objects can roundtrip"""
    # https://github.com/jsonpickle/jsonpickle/issues/362
    ll = ListLike()
    ll.internal_list.append(1)
    roundtrip_ll = jsonpickle.decode(jsonpickle.encode(ll))
    assert len(roundtrip_ll.internal_list) == len(ll.internal_list)


def test_v1_decode():
    # TODO: Find a simple example that reproduces #364
    assert True


def test_depth_tracking(pickler):
    liszt = []
    pickler.flatten([liszt, liszt, liszt, liszt, liszt])
    assert pickler._depth == -1


def test_readonly_attrs():
    """Objects with readonly attributes can roundtrip"""
    safe_str = SafeString('test')
    pickled = jsonpickle.encode(safe_str, handle_readonly=True)
    pickled_json_dict = jsonpickle.backend.json.loads(pickled)
    # make sure it's giving concise output by erroring if it includes
    # a default method which is unnecessary
    assert 'join' not in pickled_json_dict
    unpickled = jsonpickle.decode(pickled)
    assert SafeString == unpickled.__class__
    assert safe_str == unpickled


def test_readonly_str_attrs():
    """Objects with readonly string attributes can roundtrip"""
    safe_str = SafeString('test')
    # We'll first try setting handle_readonly=True when encoding.
    encoded = jsonpickle.encode(safe_str, handle_readonly=True)
    actual = jsonpickle.decode(encoded, handle_readonly=True)
    assert safe_str == actual
    # Next we'll ensure that we can decode a payload that contains readonly attributes
    # by omitting the handle_readonly option when pickling.
    encoded = jsonpickle.encode(safe_str)
    actual = jsonpickle.decode(encoded, handle_readonly=True)
    assert safe_str == actual


class PicklableNamedTuple:
    """A namedtuple wrapper that uses ``__getnewargs__``

    Demonstrates the need for protocol 2 compatibility.
    This is contrived in its use of new but it demonstrates the issue.
    """

    def __new__(cls, propnames, vals):
        # it's necessary to use the correct class name for class resolution
        # classes that fake their own names may never be unpicklable
        ntuple = collections.namedtuple(cls.__name__, propnames)
        ntuple.__getnewargs__ = lambda self: (propnames, vals)
        instance = ntuple.__new__(ntuple, *vals)
        return instance


class PicklableNamedTupleEx:
    """A namedtuple wrapper that uses ``__getnewargs__`` and ``__getnewargs_ex__``

    Demonstrates the need for protocol 4 compatibility.
    This is contrived in its use of new but it demonstrates the issue.
    """

    def __getnewargs__(self):
        raise NotImplementedError("This class needs __getnewargs_ex__")

    def __new__(cls, newargs=__getnewargs__, **kwargs):
        # it's necessary to use the correct class name for class resolution
        # classes that fake their own names may never be unpicklable
        ntuple = collections.namedtuple(cls.__name__, sorted(kwargs.keys()))
        ntuple.__getnewargs_ex__ = lambda self: ((), kwargs)
        ntuple.__getnewargs__ = newargs
        instance = ntuple.__new__(ntuple, *[b for a, b in sorted(kwargs.items())])
        return instance


class PickleProtocol2Thing:
    """An object that implements ``__getnewargs__`` for pickle protocol v2"""

    def __init__(self, *args):
        self.args = args

    def __getnewargs__(self):
        return self.args

    def __eq__(self, other):
        """
        Make PickleProtocol2Thing('slotmagic') ==
             PickleProtocol2Thing('slotmagic')
        """
        if self.__dict__ == other.__dict__ and dir(self) == dir(other):
            for prop in dir(self):
                selfprop = getattr(self, prop)
                if not callable(selfprop) and prop[0] != '_':
                    if selfprop != getattr(other, prop):
                        return False
            return True
        else:
            return False


# these two instances are used below and in tests
slotmagic = PickleProtocol2Thing('slotmagic')
dictmagic = PickleProtocol2Thing('dictmagic')


class PickleProtocol2GetState(PickleProtocol2Thing):
    """An object whose ``__getstate__`` returns a string"""

    def __new__(cls, *args):
        instance = super().__new__(cls)
        instance.newargs = args
        return instance

    def __getstate__(self):
        return 'I am magic'


class PickleProtocol2GetStateDict(PickleProtocol2Thing):
    """An object whose ``__getstate__`` returns a regular dict"""

    def __getstate__(self):
        return {'magic': True}


class PickleProtocol2GetStateNestedDict(PickleProtocol2Thing):
    """An object whose ``__getstate__`` returns a nested dict"""

    def __getstate__(self):
        return {'nested': {'magic': True}}


class PickleProtocol2GetStateSlots(PickleProtocol2Thing):
    """An object whose ``__getstate__`` return a tuple containing None and slot data"""

    def __getstate__(self):
        return (None, {'slotmagic': slotmagic})


class PickleProtocol2GetStateSlotsDict(PickleProtocol2Thing):
    """An object whose ``__getstate__`` return a tuple containing dict and slot data"""

    def __getstate__(self):
        return ({'dictmagic': dictmagic}, {'slotmagic': slotmagic})


class PickleProtocol2GetSetState(PickleProtocol2GetState):
    """An object whose ``__setstate__`` detects specific strings"""

    def __setstate__(self, state):
        """Contrived example, easy to test"""
        if state == 'I am magic':
            self.magic = True
        else:
            self.magic = False


class PickleProtocol2ChildThing:
    """Provides ``__getnewargs__`` for pickle protocol v2"""

    def __init__(self, child):
        self.child = child

    def __getnewargs__(self):
        return ([self.child],)


class PickleProtocol2ReduceString:
    """A reducible object that implements ``__reduce__`` only"""

    def __reduce__(self):
        return __name__ + '.slotmagic'


class PickleProtocol2ReduceExString:
    """A reducible object that implements both ``__reduce_ex__`` and ``__reduce__``"""

    def __reduce_ex__(self, n):
        return __name__ + '.slotmagic'

    def __reduce__(self):
        """This method is ignored"""
        assert False, 'Should not be here'


class PickleProtocol2ReduceTuple:
    """A reducible object that returns the PickleProtocol2ReduceTuple callable"""

    def __init__(self, argval, optional=None):
        self.argval = argval
        self.optional = optional

    def __reduce__(self):
        return (
            PickleProtocol2ReduceTuple,  # callable
            ('yam', 1),  # args
            None,  # state
            iter([]),  # listitems
            iter([]),  # dictitems
        )


class ReducibleIterator:
    """A reducible iterator"""

    def __next__(self):
        raise StopIteration()

    def __iter__(self):
        return self

    def __reduce__(self):
        return ReducibleIterator, ()


def protocol_2_reduce_tuple_func(*args):
    """Return a pickle protocol v2 reducer"""
    return PickleProtocol2ReduceTupleFunc(*args)


class PickleProtocol2ReduceTupleFunc:
    """A reducible object with args and optional state"""

    def __init__(self, argval, optional=None):
        self.argval = argval
        self.optional = optional

    def __reduce__(self):
        return (
            protocol_2_reduce_tuple_func,  # callable
            ('yam', 1),  # args
            None,  # state
            iter([]),  # listitems
            iter([]),  # dictitems
        )


def __newobj__(lol, fail):
    """newobj is special-cased so that it is not actually called"""


class PickleProtocol2ReduceNewobj(PickleProtocol2ReduceTupleFunc):
    """Reducible object that return the ``__newobj__`` callable and args"""

    def __new__(cls, *args):
        inst = super(cls, cls).__new__(cls)
        inst.newargs = args
        return inst

    def __reduce__(self):
        return (
            __newobj__,  # callable
            (PickleProtocol2ReduceNewobj, 'yam', 1),  # args
            None,  # state
            iter([]),  # listitems
            iter([]),  # dictitems
        )


class PickleProtocol2ReduceTupleState(PickleProtocol2ReduceTuple):
    """Reducible object with args and state"""

    def __reduce__(self):
        return (
            PickleProtocol2ReduceTuple,  # callable
            ('yam', 1),  # args
            {'foo': 1},  # state
            iter([]),  # listitems
            iter([]),  # dictitems
        )


class PickleProtocol2ReduceTupleSetState(PickleProtocol2ReduceTuple):
    """Reducible object with args and state that implements ``__setstate__``"""

    def __reduce__(self):
        return (
            type(self),  # callable
            ('yam', 1),  # args
            {'foo': 1},  # state
            iter([]),  # listitems
            iter([]),  # dictitems
        )

    def __setstate__(self, state):
        self.bar = state['foo']


class PickleProtocol2ReduceTupleStateSlots:
    """Reducible object with tuple ``__slots__``"""

    __slots__ = ('argval', 'optional', 'foo')

    def __init__(self, argval, optional=None):
        self.argval = argval
        self.optional = optional

    def __reduce__(self):
        return (
            PickleProtocol2ReduceTuple,  # callable
            ('yam', 1),  # args
            {'foo': 1},  # state
            iter([]),  # listitems
            iter([]),  # dictitems
        )


class PickleProtocol2ReduceListitemsAppend:
    """A reducible object that only implements append()"""

    def __init__(self):
        self.inner = []

    def append(self, item):
        self.inner.append(item)

    def __reduce__(self):
        return (
            PickleProtocol2ReduceListitemsAppend,  # callable
            (),  # args
            {},  # state
            iter(['foo', 'bar']),  # listitems
            iter([]),  # dictitems
        )


class PickleProtocol2ReduceListitemsExtend:
    """A reducible object that only implements extend()"""

    def __init__(self):
        self.inner = []

    def extend(self, items):
        self.inner.exend(items)

    def __reduce__(self):
        return (
            PickleProtocol2ReduceListitemsAppend,  # callable
            (),  # args
            {},  # state
            iter(['foo', 'bar']),  # listitems
            iter([]),  # dictitems
        )


class PickleProtocol2ReduceDictitems:
    """A reducible object with dictitems that implements ``__setitem__``"""

    def __init__(self):
        self.inner = {}

    def __setitem__(self, k, v):
        return self.inner.__setitem__(k, v)

    def __reduce__(self):
        return (
            PickleProtocol2ReduceDictitems,  # callable
            (),  # args
            {},  # state
            [],  # listitems
            iter(zip(['foo', 'bar'], ['foo', 'bar'])),  # dictitems
        )


def test_pickle_newargs_ex():
    """
    Ensure we can pickle and unpickle an object whose class needs arguments
    to __new__ and get back the same type
    """
    instance = PicklableNamedTupleEx(**{'a': 'b', 'n': 2})
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert instance == decoded


def test_validate_reconstruct_by_newargs_ex():
    """
    Ensure that the exemplar tuple's __getnewargs_ex__ works
    This is necessary to know whether the breakage exists
    in jsonpickle or not
    """
    instance = PicklableNamedTupleEx(**{'a': 'b', 'n': 2})
    args, kwargs = instance.__getnewargs_ex__()
    newinstance = PicklableNamedTupleEx.__new__(PicklableNamedTupleEx, *args, **kwargs)
    assert instance == newinstance


def test_references_named_tuple():
    """Object references and identities are preserved inside a named tuple"""
    shared = Thing('shared')
    instance = PicklableNamedTupleEx(**{'a': shared, 'n': shared})
    child = Thing('child')
    shared.child = child
    child.child = instance
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded[0] == decoded[1]
    assert decoded[0] is decoded[1]
    assert decoded.a is decoded.n
    assert decoded.a.name == 'shared'
    assert decoded.a.child.name == 'child'
    assert decoded.a.child.child is decoded
    assert decoded.n.child.child is decoded
    assert decoded.a.child is decoded.n.child
    assert decoded.__class__.__name__ == PicklableNamedTupleEx.__name__
    # TODO the class itself looks just like the real class, but it's
    # actually a reconstruction; PicklableNamedTupleEx is not type(decoded).
    assert decoded.__class__ is not PicklableNamedTupleEx


def test_reduce_complex_num():
    """Test complex/imaginary numbers"""
    instance = 5j
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded == instance


def test_reduce_complex_zero():
    """Test complex/imaginary zero"""
    instance = 0j
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded == instance


def test_reduce_dictitems():
    """Test reduce with dictitems set (as a generator)"""
    instance = PickleProtocol2ReduceDictitems()
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.inner == {'foo': 'foo', 'bar': 'bar'}


def test_reduce_listitems_extend():
    """Test reduce with listitems set (as a generator), yielding single items"""
    instance = PickleProtocol2ReduceListitemsExtend()
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.inner == ['foo', 'bar']


def test_reduce_listitems_append():
    """Test reduce with listitems set (as a generator), yielding single items"""
    instance = PickleProtocol2ReduceListitemsAppend()
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.inner == ['foo', 'bar']


def test_reduce_state_setstate():
    """Objects with ``__setstate__`` and optional state arguments can roundtrip"""
    instance = PickleProtocol2ReduceTupleSetState(5)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.argval == 'yam'
    assert decoded.optional == 1
    assert decoded.bar == 1
    assert not hasattr(decoded, 'foo')


def test_reduce_state_no_dict():
    """Objects with state but without `__dict__`` and ``__setstate__`` can roundtrip"""
    instance = PickleProtocol2ReduceTupleStateSlots(5)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.argval == 'yam'
    assert decoded.optional == 1
    assert decoded.foo == 1


def test_reduce_state_dict():
    """Reduce an object with __dict__ and no __setstate__"""
    instance = PickleProtocol2ReduceTupleState(5)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.argval == 'yam'
    assert decoded.optional == 1
    assert decoded.foo == 1


def test_reduce_basic():
    """Reduce an object with callable and args only"""
    instance = PickleProtocol2ReduceTuple(5)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.argval == 'yam'
    assert decoded.optional == 1


def test_reduce_basic_func():
    """Reduce an object with args and a module-level callable"""
    instance = PickleProtocol2ReduceTupleFunc(5)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.argval == 'yam'
    assert decoded.optional == 1


def test_reduce_newobj():
    """Reduce an object with a callable __newobj__"""
    instance = PickleProtocol2ReduceNewobj(5)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.newargs == ('yam', 1)


def test_reduce_iter():
    """Iterators with ``__reduce__`` can roundtrip"""
    instance = iter('123')
    assert util.is_iterator(instance)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert next(decoded) == '1'
    assert next(decoded) == '2'
    assert next(decoded) == '3'


def test_reduce_iterable():
    """Reducible objects that are iterable should also pickle"""
    instance = ReducibleIterator()
    assert util.is_iterator(instance)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert isinstance(decoded, ReducibleIterator)


def test_reduce_string():
    """Handle redirection to another object when ``__reduce__`` returns a string"""
    instance = PickleProtocol2ReduceString()
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded == slotmagic


def test_reduce_ex_string():
    """Handle redirection to another object when ``__reduce_ex__`` returns a string

    Ensure that ``__reduce_ex__`` has higher priority than ``__reduce__``.
    """
    instance = PickleProtocol2ReduceExString()
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded == slotmagic


def test_pickle_newargs():
    """Objects whose class needs arguments to ``__new__`` can roundtrip"""
    instance = PicklableNamedTuple(('a', 'b'), (1, 2))
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert instance == decoded


def test_validate_reconstruct_by_newargs():
    """Validate PicklableNamedTuple's Python semantics

    Ensure that the exemplar tuple's __getnewargs__ works This is necessary to
    know whether potential breakage exists in jsonpickle or in the test class.
    """
    instance = PicklableNamedTuple(('a', 'b'), (1, 2))
    newinstance = PicklableNamedTuple.__new__(
        PicklableNamedTuple, *(instance.__getnewargs__())
    )
    assert instance == newinstance


def test_getnewargs_priority():
    """newargs must be used before py/state when decoding

    As per PEP 307, classes are not supposed to implement
    all three magic methods.
    """
    instance = PickleProtocol2GetState('whatevs')
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.newargs == ('whatevs',)


def test_restore_dict_state():
    """getstate without setstate can roundtrip

    If ``__getstate__`` returns a dict and there is no custom ``__setstate__`` then the
    dict is used as a source of variables to restore.
    """
    instance = PickleProtocol2GetStateDict('whatevs')
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.magic


def test_restore_nested_dict_state_with_references_preserved():
    """Nested dicts in "py/state" retain references and are not duplicated"""
    instance1 = PickleProtocol2GetStateNestedDict('whatevs')
    instance2 = PickleProtocol2GetStateNestedDict('different')
    encoded = jsonpickle.encode([instance1, instance1, instance2, instance2])
    decoded = jsonpickle.decode(encoded)
    assert decoded[0].nested['magic']
    assert decoded[1] is decoded[0]
    assert decoded[2].nested['magic']
    assert decoded[3] is decoded[2]


def test_restore_slots_state():
    """Serialize objects with __slots__

    Ensure that if getstate returns a 2-tuple with a dict in the second
    position, and there is no custom __setstate__, the dict is used as a
    source of variables to restore.
    """
    instance = PickleProtocol2GetStateSlots('whatevs')
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.slotmagic.__dict__ == slotmagic.__dict__
    assert decoded.slotmagic == slotmagic


def test_restore_slots_dict_state():
    """getstate with dicts in both positions can roundtrip

    If ``__getstate__`` returns a 2-tuple with a dict in both positions,
    and there is no custom ``__setstate__``, the dicts are used as a source of
    variables to restore.
    """
    instance = PickleProtocol2GetStateSlotsDict('whatevs')
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert PickleProtocol2Thing('slotmagic') == PickleProtocol2Thing('slotmagic')
    assert decoded.slotmagic.__dict__ == slotmagic.__dict__
    assert decoded.slotmagic == slotmagic
    assert decoded.dictmagic == dictmagic


def test_setstate():
    """Ensure output of getstate is passed to setstate"""
    instance = PickleProtocol2GetSetState('whatevs')
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded.magic


def test_handles_nested_objects():
    """Nested objects can roundtrip"""
    child = PickleProtocol2Thing(None)
    instance = PickleProtocol2Thing(child, child)
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert PickleProtocol2Thing == decoded.__class__
    assert PickleProtocol2Thing == decoded.args[0].__class__
    assert PickleProtocol2Thing == decoded.args[1].__class__
    assert decoded.args[0] is decoded.args[1]


def test_cyclical_objects():
    """Cyclical objects with inner tuples and lists can roundtrip"""
    _test_cyclical_objects(True)
    _test_cyclical_objects(False)


def _test_cyclical_objects(use_tuple):
    child = Capture(None)
    instance = Capture(child, child)
    # create a cycle
    if use_tuple:
        child.args = (instance,)
    else:
        child.args = [instance]
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    # Ensure the right objects were constructed
    assert Capture == decoded.__class__
    assert Capture == decoded.args[0].__class__
    assert Capture == decoded.args[1].__class__
    assert Capture == decoded.args[0].args[0].__class__
    assert Capture == decoded.args[1].args[0].__class__
    # It's turtles all the way down
    assert (
        Capture
        == decoded.args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .args[0]
        .__class__
    )
    # Ensure that references are properly constructed
    assert decoded.args[0] is decoded.args[1]
    assert decoded is decoded.args[0].args[0]
    assert decoded is decoded.args[1].args[0]
    assert decoded.args[0] is decoded.args[0].args[0].args[0]
    assert decoded.args[0] is decoded.args[1].args[0].args[0]


def test_handles_cyclical_objects_in_lists():
    """Cyclical objects in lists can roundtrip"""
    child = PickleProtocol2ChildThing(None)
    instance = PickleProtocol2ChildThing([child, child])
    child.child = instance  # create a cycle
    encoded = jsonpickle.encode(instance)
    decoded = jsonpickle.decode(encoded)
    assert decoded is decoded.child[0].child
    assert decoded is decoded.child[1].child


def test_cyclical_objects_unpickleable():
    """Cyclical objects can roundtrip with their references retained"""
    _test_cyclical_objects_unpickleable(True)
    _test_cyclical_objects_unpickleable(False)


def _test_cyclical_objects_unpickleable(use_tuple):
    child = Capture(None)
    instance = Capture(child, child)
    # create a cycle
    if use_tuple:
        child.args = (instance,)
    else:
        child.args = [instance]
    encoded = jsonpickle.encode(instance, unpicklable=False)
    decoded = jsonpickle.decode(encoded)
    assert isinstance(decoded, dict)
    assert 'args' in decoded
    assert 'kwargs' in decoded
    # Tuple is lost via json
    args = decoded['args']
    assert isinstance(args, list)
    # Get the children
    assert len(args) == 2
    decoded_child0 = args[0]
    decoded_child1 = args[1]
    # Circular references become None
    assert decoded_child0 == {'args': [None], 'kwargs': {}}
    assert decoded_child1 == {'args': [None], 'kwargs': {}}


def test_dict_references_are_preserved():
    """References and object identities are preserved in dicts"""
    data = {}
    actual = jsonpickle.decode(jsonpickle.encode([data, data]))
    assert isinstance(actual, list)
    assert isinstance(actual[0], dict)
    assert isinstance(actual[1], dict)
    assert actual[0] is actual[1]


def test_repeat_objects_are_expanded():
    """All objects are present in the json output"""
    # When references are disabled we should create expanded copies
    # of any object that appears more than once in the object stream.
    alice = Thing('alice')
    bob = Thing('bob')
    alice.child = bob
    car = Thing('car')
    car.driver = alice
    car.owner = alice
    car.passengers = [alice, bob]
    pickler = jsonpickle.Pickler(make_refs=False)
    flattened = pickler.flatten(car)
    assert flattened['name'] == 'car'
    assert flattened['driver']['name'] == 'alice'
    assert flattened['owner']['name'] == 'alice'
    assert flattened['passengers'][0]['name'] == 'alice'
    assert flattened['passengers'][1]['name'] == 'bob'
    assert flattened['driver']['child']['name'] == 'bob'
    assert flattened['passengers'][0]['child']['name'] == 'bob'



import os
import sys

try:
    import furo
except ImportError:
    furo = None
try:
    import rst.linker as rst_linker
except ImportError:
    rst_linker = None

# Use the source tree.
sys.path.insert(1, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
]

project = 'jsonpickle'
master_doc = 'index'

if furo is not None:
    html_theme = 'furo'

# Link dates and other references in the changelog
if rst_linker is not None:
    extensions += ['rst.linker']

package_url = 'https://github.com/jsonpickle/jsonpickle'
link_files = {
    '../CHANGES.rst': dict(
        using=dict(GH='https://github.com'),
        replace=[
            dict(
                pattern=r'(Issue #|\B#)(?P<issue>\d+)',
                url=package_url + '/issues/{issue}',
            ),
            dict(
                pattern=r'\B\+(?P<pull>\d+)',
                url=package_url + '/pull/{pull}',
            ),
            dict(
                pattern=r'(?m:^((?P<scm_version>v?\d+(\.\d+){1,2}))\n[-=]+\n)',
                with_scm='{text}\n{rev[timestamp]:%d %b %Y}\n',
            ),
            dict(
                pattern=r'PEP[- ](?P<pep_number>\d+)',
                url='https://www.python.org/dev/peps/pep-{pep_number:0>4}/',
            ),
        ],
    )
}

# Be strict about any broken references
nitpicky = True

sphinx_disable = os.environ.get('JSONPICKLE_SPHINX_DISABLE', '')
if 'intersphinx' not in sphinx_disable:
    extensions += ['sphinx.ext.intersphinx']
intersphinx_mapping = {
    'python': ('https://docs.python.org/3', None),
    'sphinx': ('https://www.sphinx-doc.org/en/stable/', None),
}

# Preserve authored syntax for defaults
autodoc_preserve_defaults = True



import array
import collections
import datetime
import decimal
import enum
import queue
import re
import threading

import pytest
from helper import SkippableTest

import jsonpickle
from jsonpickle import handlers, tags, util


@pytest.fixture
def pickler():
    """Returns a default-constructed pickler"""
    return jsonpickle.pickler.Pickler()


@pytest.fixture
def unpickler():
    """Returns a default-constructed unpickler"""
    return jsonpickle.unpickler.Unpickler()


class Thing:
    def __init__(self, name):
        self.name = name
        self.child = None


class DictSubclass(dict):
    name = 'Test'


class ListSubclass(list):
    pass


class BrokenReprThing(Thing):
    def __repr__(self):
        raise Exception('%s has a broken repr' % self.name)

    def __str__(self):
        return '<BrokenReprThing "%s">' % self.name


class GetstateDict(dict):
    def __init__(self, name, **kwargs):
        dict.__init__(self, **kwargs)
        self.name = name
        self.active = False

    def __getstate__(self):
        return (self.name, dict(self.items()))

    def __setstate__(self, state):
        self.name, vals = state
        self.update(vals)
        self.active = True


class GetstateOnly:
    def __init__(self, a=1, b=2):
        self.a = a
        self.b = b

    def __getstate__(self):
        return [self.a, self.b]


class GetstateReturnsList:
    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __getstate__(self):
        return [self.x, self.y]

    def __setstate__(self, state):
        self.x, self.y = state[0], state[1]


class GetstateRecursesInfintely:
    def __getstate__(self):
        return GetstateRecursesInfintely()


class ListSubclassWithInit(list):
    def __init__(self, attr):
        self.attr = attr
        super().__init__()


NamedTuple = collections.namedtuple('NamedTuple', 'a, b, c')


class ObjWithJsonPickleRepr:
    def __init__(self):
        self.data = {'a': self}

    def __repr__(self):
        return jsonpickle.encode(self)


class OldStyleClass:
    pass


class SetSubclass(set):
    pass


def func(x):
    return x


class ThingWithFunctionRefs:
    def __init__(self):
        self.fn = func


class ThingWithQueue:
    def __init__(self):
        self.child_1 = queue.Queue()
        self.child_2 = queue.Queue()
        self.childref_1 = self.child_1
        self.childref_2 = self.child_2


class ThingWithSlots:
    __slots__ = ('a', 'b')

    def __init__(self, a, b):
        self.a = a
        self.b = b


class ThingWithInheritedSlots(ThingWithSlots):
    __slots__ = ('c',)

    def __init__(self, a, b, c):
        ThingWithSlots.__init__(self, a, b)
        self.c = c


class ThingWithIterableSlots:
    __slots__ = iter('ab')

    def __init__(self, a, b):
        self.a = a
        self.b = b


class ThingWithStringSlots:
    __slots__ = 'ab'

    def __init__(self, a, b):
        self.ab = a + b


class ThingWithSelfAsDefaultFactory(collections.defaultdict):
    """defaultdict subclass that uses itself as its default factory"""

    def __init__(self):
        self.default_factory = self

    def __call__(self):
        return self.__class__()


class ThingWithClassAsDefaultFactory(collections.defaultdict):
    """defaultdict subclass that uses its class as its default factory"""

    def __init__(self):
        self.default_factory = self.__class__

    def __call__(self):
        return self.__class__()


class ThingWithDefaultFactoryRegistry:
    """Counts calls to ThingWithDefaultFactory.__init__()"""

    count = 0  # Incremented by ThingWithDefaultFactory().


class ThingWithDefaultFactory:
    """Class that ensures that provides a default factory"""

    def __new__(cls, *args, **kwargs):
        """Ensure that jsonpickle uses our constructor during initialization"""
        return super().__new__(cls, *args, **kwargs)

    def __init__(self):
        self.default_factory = self.__class__
        # Keep track of how many times this constructor has been run.
        ThingWithDefaultFactoryRegistry.count += 1


class IntEnumTest(enum.IntEnum):
    X = 1
    Y = 2


class StringEnumTest(enum.Enum):
    A = 'a'
    B = 'b'


class SubEnum(enum.Enum):
    a = 1
    b = 2


class EnumClass:
    def __init__(self):
        self.enum_a = SubEnum.a
        self.enum_b = SubEnum.b


class MessageTypes(enum.Enum):
    STATUS = ('STATUS',)
    CONTROL = 'CONTROL'


class MessageStatus(enum.Enum):
    OK = ('OK',)
    ERROR = 'ERROR'


class MessageCommands(enum.Enum):
    STATUS_ALL = 'STATUS_ALL'


class Message:
    def __init__(self, message_type, command, status=None, body=None):
        self.message_type = MessageTypes(message_type)
        if command:
            self.command = MessageCommands(command)
        if status:
            self.status = MessageStatus(status)
        if body:
            self.body = body


class ThingWithTimedeltaAttribute:
    def __init__(self, offset):
        self.offset = datetime.timedelta(offset)

    def __getinitargs__(self):
        return (self.offset,)


class FailSafeTestCase(SkippableTest):
    class BadClass:
        def __getstate__(self):
            raise ValueError('Intentional error')

    good = 'good'

    to_pickle = [BadClass(), good]

    def test_no_error(self):
        encoded = jsonpickle.encode(self.to_pickle, fail_safe=lambda e: None)
        decoded = jsonpickle.decode(encoded)
        assert decoded[0] is None
        assert decoded[1] == 'good'

    def test_error_recorded(self):
        exceptions = []

        def recorder(exception):
            exceptions.append(exception)

        jsonpickle.encode(self.to_pickle, fail_safe=recorder)
        assert len(exceptions) == 1
        assert isinstance(exceptions[0], Exception)

    def test_custom_err_msg(self):
        CUSTOM_ERR_MSG = 'custom err msg'
        encoded = jsonpickle.encode(self.to_pickle, fail_safe=lambda e: CUSTOM_ERR_MSG)
        decoded = jsonpickle.decode(encoded)
        assert decoded[0] == CUSTOM_ERR_MSG


class IntKeysObject:
    def __init__(self):
        self.data = {0: 0}

    def __getstate__(self):
        return self.__dict__


class ExceptionWithArguments(Exception):
    def __init__(self, value):
        super().__init__('test')
        self.value = value


class ThingWithExclusion:
    _jsonpickle_exclude = ["foo"]

    def __init__(self, a):
        self.foo = 1
        self.bar = a


class ThingWithExcludeSubclass:
    def __init__(self, foo):
        self.foo = foo
        self.thing = ThingWithExclusion(3)


def test_list_subclass(pickler, unpickler):
    obj = ListSubclass()
    obj.extend([1, 2, 3])
    flattened = pickler.flatten(obj)
    assert tags.OBJECT in flattened
    assert tags.SEQ in flattened
    assert len(flattened[tags.SEQ]) == 3
    for v in obj:
        assert v in flattened[tags.SEQ]
    restored = unpickler.restore(flattened)
    assert type(restored) is ListSubclass
    assert restored == obj


def test_list_subclass_with_init(pickler, unpickler):
    obj = ListSubclassWithInit('foo')
    assert obj.attr == 'foo'
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert type(inflated) is ListSubclassWithInit


def test_list_subclass_with_data(pickler, unpickler):
    obj = ListSubclass()
    obj.extend([1, 2, 3])
    data = SetSubclass([1, 2, 3])
    obj.data = data
    flattened = pickler.flatten(obj)
    restored = unpickler.restore(flattened)
    assert restored == obj
    assert type(restored.data) is SetSubclass
    assert restored.data == data


def test_set_subclass(pickler, unpickler):
    obj = SetSubclass([1, 2, 3])
    flattened = pickler.flatten(obj)
    assert tags.OBJECT in flattened
    assert tags.SEQ in flattened
    assert len(flattened[tags.SEQ]) == 3
    for v in obj:
        assert v in flattened[tags.SEQ]
    restored = unpickler.restore(flattened)
    assert type(restored) is SetSubclass
    assert restored == obj


def test_set_subclass_with_data(pickler, unpickler):
    obj = SetSubclass([1, 2, 3])
    data = ListSubclass()
    data.extend([1, 2, 3])
    obj.data = data
    flattened = pickler.flatten(obj)
    restored = unpickler.restore(flattened)
    assert restored.data.__class__ == ListSubclass
    assert restored.data == data


def test_decimal(pickler, unpickler):
    obj = decimal.Decimal('0.5')
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert isinstance(inflated, decimal.Decimal)


def test_oldstyleclass(pickler, unpickler):
    obj = OldStyleClass()
    obj.value = 1234
    flattened = pickler.flatten(obj)
    assert flattened['value'] == 1234
    inflated = unpickler.restore(flattened)
    assert inflated.value == 1234


def test_dictsubclass(pickler, unpickler):
    obj = DictSubclass()
    obj['key1'] = 1
    expect = {
        tags.OBJECT: 'object_test.DictSubclass',
        'key1': 1,
        '__dict__': {},
    }
    flattened = pickler.flatten(obj)
    assert expect == flattened
    inflated = unpickler.restore(flattened)
    assert type(inflated) is DictSubclass
    assert inflated['key1'] == 1
    assert inflated.name == 'Test'


def test_dictsubclass_notunpickable(pickler, unpickler):
    pickler.unpicklable = False
    obj = DictSubclass()
    obj['key1'] = 1
    flattened = pickler.flatten(obj)
    assert flattened['key1'] == 1
    assert tags.OBJECT not in flattened
    inflated = unpickler.restore(flattened)
    assert inflated['key1'] == 1


def test_getstate_dict_subclass_structure(pickler):
    obj = GetstateDict('test')
    obj['key1'] = 1
    flattened = pickler.flatten(obj)
    assert tags.OBJECT in flattened
    assert 'object_test.GetstateDict' == flattened[tags.OBJECT]
    assert tags.STATE in flattened
    assert tags.TUPLE in flattened[tags.STATE]
    assert ['test' == {'key1': 1}], flattened[tags.STATE][tags.TUPLE]


def test_getstate_dict_subclass_roundtrip_simple(pickler, unpickler):
    obj = GetstateDict('test')
    obj['key1'] = 1
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert inflated['key1'] == 1
    assert inflated.name == 'test'


def test_getstate_dict_subclass_roundtrip_cyclical(pickler, unpickler):
    obj = GetstateDict('test')
    obj['key1'] = 1
    # The "name" field of obj2 points to obj (reference)
    obj2 = GetstateDict(obj)
    # The "obj2" key in obj points to obj2 (cyclical reference)
    obj['obj2'] = obj2
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    # The dict must be preserved
    assert inflated['key1'] == 1
    # __getstate__/__setstate__ must have been run
    assert inflated.name == 'test'
    assert inflated.active is True
    assert inflated['obj2'].active is True
    # The reference must be preserved
    assert inflated is inflated['obj2'].name


def test_getstate_list_simple(pickler, unpickler):
    obj = GetstateReturnsList(1, 2)
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert inflated.x == 1
    assert inflated.y == 2


def test_getstate_list_inside_list(pickler, unpickler):
    obj1 = GetstateReturnsList(1, 2)
    obj2 = GetstateReturnsList(3, 4)
    obj = [obj1, obj2]
    flattened = pickler.flatten(obj)
    inflated = unpickler.restore(flattened)
    assert inflated[0].x == 1
    assert inflated[0].y == 2
    assert inflated[1].x == 3
    assert inflated[1].y == 4


def test_getstate_with_getstate_only(pickler, unpickler):
    obj = GetstateOnly()
    a = obj.a = 'this object implements'
    b = obj.b = '__getstate__ but not __setstate__'
    expect = [a, b]
    flat = pickler.flatten(obj)
    actual = flat[tags.STATE]
    assert expect == actual
    restored = unpickler.restore(flat)
    assert expect == restored


def test_thing_with_queue(pickler, unpickler):
    obj = ThingWithQueue()
    flattened = pickler.flatten(obj)
    restored = unpickler.restore(flattened)
    assert type(restored.child_1) is type(queue.Queue())
    assert type(restored.child_2) is type(queue.Queue())
    # Check references
    assert restored.child_1 is restored.childref_1
    assert restored.child_2 is restored.childref_2


def test_thing_with_func(pickler, unpickler):
    obj = ThingWithFunctionRefs()
    obj.ref = obj
    flattened = pickler.flatten(obj)
    restored = unpickler.restore(flattened)
    assert restored.fn is obj.fn
    expect = 'success'
    actual1 = restored.fn(expect)
    assert expect == actual1
    assert restored is restored.ref


def test_thing_with_compiled_regex(pickler, unpickler):
    rgx = re.compile(r'(.*)(cat)')
    obj = Thing(rgx)
    flattened = pickler.flatten(obj)
    restored = unpickler.restore(flattened)
    match = restored.name.match('fatcat')
    assert 'fat' == match.group(1)
    assert 'cat' == match.group(2)


def test_base_object_roundrip(pickler, unpickler):
    roundtrip = unpickler.restore(pickler.flatten(object()))
    assert type(roundtrip) is object


def test_enum34(pickler, unpickler):
    restore = unpickler.restore
    flatten = pickler.flatten

    def roundtrip(obj):
        return restore(flatten(obj))

    assert roundtrip(IntEnumTest.X) is IntEnumTest.X
    assert roundtrip(IntEnumTest) is IntEnumTest
    assert roundtrip(StringEnumTest.A) is StringEnumTest.A
    assert roundtrip(StringEnumTest) is StringEnumTest


def test_bytes_unicode(pickler, unpickler):
    b1 = b'foo'
    b2 = b'foo\xff'
    u1 = 'foo'
    # unicode strings get encoded/decoded as is
    encoded = pickler.flatten(u1)
    assert encoded == u1
    assert isinstance(encoded, str)
    decoded = unpickler.restore(encoded)
    assert decoded == u1
    assert isinstance(decoded, str)
    # bytestrings are wrapped in py 3
    encoded = pickler.flatten(b1)
    assert encoded != u1
    encoded_ustr = util.b64encode(b'foo')
    assert {tags.B64: encoded_ustr} == encoded
    assert isinstance(encoded[tags.B64], str)
    decoded = unpickler.restore(encoded)
    assert decoded == b1
    assert isinstance(decoded, bytes)
    # bytestrings that we can't decode to UTF-8 will always be wrapped
    encoded = pickler.flatten(b2)
    assert encoded != b2
    encoded_ustr = util.b64encode(b'foo\xff')
    assert {tags.B64: encoded_ustr} == encoded
    assert isinstance(encoded[tags.B64], str)
    decoded = unpickler.restore(encoded)
    assert decoded == b2
    assert isinstance(decoded, bytes)


def test_nested_objects(pickler, unpickler):
    obj = ThingWithTimedeltaAttribute(99)
    flattened = pickler.flatten(obj)
    restored = unpickler.restore(flattened)
    assert restored.offset == datetime.timedelta(99)


def test_threading_lock(pickler, unpickler):
    obj = Thing('lock')
    obj.lock = threading.Lock()
    lock_class = obj.lock.__class__
    # Roundtrip and make sure we get a lock object.
    json = pickler.flatten(obj)
    clone = unpickler.restore(json)
    assert isinstance(clone.lock, lock_class)
    assert not clone.lock.locked()
    # Serializing a locked lock should create a locked clone.
    assert obj.lock.acquire()
    json = pickler.flatten(obj)
    obj.lock.release()
    # Restore the locked lock state.
    clone = unpickler.restore(json)
    assert clone.lock.locked()
    clone.lock.release()


def _test_array_roundtrip(pickler, unpickler, obj):
    """Roundtrip an array and test invariants"""
    json = pickler.flatten(obj)
    clone = unpickler.restore(json)
    assert isinstance(clone, array.array)
    assert obj.typecode == clone.typecode
    assert len(obj) == len(clone)
    for j, k in zip(obj, clone):
        assert j == k
    assert obj == clone


def test_array_handler_numeric(pickler, unpickler):
    """Test numeric array.array typecodes that work in Python2+3"""
    typecodes = ('b', 'B', 'h', 'H', 'i', 'I', 'l', 'L', 'f', 'd')
    for typecode in typecodes:
        obj = array.array(typecode, (1, 2, 3))
        _test_array_roundtrip(pickler, unpickler, obj)


def test_exceptions_with_arguments(pickler, unpickler):
    """Ensure that we can roundtrip Exceptions that take arguments"""
    obj = ExceptionWithArguments('example')
    json = pickler.flatten(obj)
    clone = unpickler.restore(json)
    assert obj.value == clone.value
    assert obj.args == clone.args


def test_getstate_does_not_recurse_infinitely():
    """Serialize an object with a __getstate__ that recurses forever"""
    obj = GetstateRecursesInfintely()
    pickler = jsonpickle.pickler.Pickler(max_depth=5)
    actual = pickler.flatten(obj)
    assert isinstance(actual, dict)
    assert tags.OBJECT in actual


def test_defaultdict_roundtrip():
    """Make sure we can handle collections.defaultdict(list)"""
    # setup
    defaultdict = collections.defaultdict
    defdict = defaultdict(list)
    defdict['a'] = 1
    defdict['b'].append(2)
    defdict['c'] = defaultdict(dict)
    # jsonpickle work your magic
    encoded = jsonpickle.encode(defdict)
    newdefdict = jsonpickle.decode(encoded)
    # jsonpickle never fails
    assert newdefdict['a'] == 1
    assert newdefdict['b'] == [2]
    assert type(newdefdict['c']) is defaultdict
    assert defdict.default_factory == list
    assert newdefdict.default_factory == list


def test_default_factory_initialization():
    """Ensure that ThingWithDefaultFactory()'s constructor is called"""
    thing = ThingWithDefaultFactory()
    assert ThingWithDefaultFactoryRegistry.count == 1

    thing.is_awesome = True
    encoded = jsonpickle.encode(thing)
    new_thing = jsonpickle.decode(encoded)

    assert ThingWithDefaultFactoryRegistry.count == 2
    assert new_thing.is_awesome


def test_defaultdict_roundtrip_simple_lambda():
    """Make sure we can handle defaultdict(lambda: defaultdict(int))"""
    # setup a sparse collections.defaultdict with simple lambdas
    defaultdict = collections.defaultdict
    defdict = defaultdict(lambda: defaultdict(int))
    defdict[0] = 'zero'
    defdict[1] = defaultdict(lambda: defaultdict(dict))
    defdict[1][0] = 'zero'
    # roundtrip
    encoded = jsonpickle.encode(defdict, keys=True)
    newdefdict = jsonpickle.decode(encoded, keys=True)
    assert newdefdict[0] == 'zero'
    assert type(newdefdict[1]) is defaultdict
    assert newdefdict[1][0] == 'zero'
    assert newdefdict[1][1] == {}  # inner defaultdict
    assert newdefdict[2][0] == 0  # outer defaultdict
    assert type(newdefdict[3]) is defaultdict
    # outer-most defaultdict
    assert newdefdict[3].default_factory == int


def test_defaultdict_roundtrip_simple_lambda2():
    """Serialize a defaultdict that contains a lambda"""
    defaultdict = collections.defaultdict
    payload = {'a': defaultdict(lambda: 0)}
    defdict = defaultdict(lambda: 0, payload)
    # roundtrip
    encoded = jsonpickle.encode(defdict, keys=True)
    decoded = jsonpickle.decode(encoded, keys=True)
    assert type(decoded) is defaultdict
    assert type(decoded['a']) is defaultdict


def test_defaultdict_and_things_roundtrip_simple_lambda():
    """Serialize a default dict that contains a lambda and objects"""
    thing = Thing('a')
    defaultdict = collections.defaultdict
    defdict = defaultdict(lambda: 0)
    obj = [defdict, thing, thing]
    # roundtrip
    encoded = jsonpickle.encode(obj, keys=True)
    decoded = jsonpickle.decode(encoded, keys=True)
    assert decoded[0].default_factory() == 0
    assert decoded[1] is decoded[2]


def _test_defaultdict_tree(tree, cls):
    tree['A']['B'] = 1
    tree['A']['C'] = 2
    # roundtrip
    encoded = jsonpickle.encode(tree)
    newtree = jsonpickle.decode(encoded)
    # make sure we didn't lose anything
    assert type(newtree) is cls
    assert type(newtree['A']) is cls
    assert newtree['A']['B'] == 1
    assert newtree['A']['C'] == 2
    # ensure that the resulting default_factory is callable and creates
    # a new instance of cls.
    assert type(newtree['A'].default_factory()) == cls
    # we've never seen 'D' before so the reconstructed defaultdict tree
    # should create an instance of cls.
    assert type(newtree['A']['D']) is cls
    # ensure that proxies do not escape into user code
    assert type(newtree.default_factory) is not jsonpickle.unpickler._Proxy
    assert type(newtree['A'].default_factory) is not jsonpickle.unpickler._Proxy
    assert type(newtree['A']['Z'].default_factory) is not jsonpickle.unpickler._Proxy
    return newtree


def test_defaultdict_subclass_with_self_as_default_factory():
    """Serialize a defaultdict subclass with self as its default factory"""
    cls = ThingWithSelfAsDefaultFactory
    tree = cls()
    newtree = _test_defaultdict_tree(tree, cls)
    assert type(newtree['A'].default_factory) is cls
    assert newtree.default_factory is newtree
    assert newtree['A'].default_factory is newtree['A']
    assert newtree['Z'].default_factory is newtree['Z']


def test_defaultdict_subclass_with_class_as_default_factory():
    """Serialize a defaultdict with a class as its default factory"""
    cls = ThingWithClassAsDefaultFactory
    tree = cls()
    newtree = _test_defaultdict_tree(tree, cls)
    assert newtree.default_factory is cls
    assert newtree['A'].default_factory is cls
    assert newtree['Z'].default_factory is cls


def test_posix_stat_result():
    """Serialize a posix.stat() result"""
    try:
        import posix
    except ImportError:
        return
    expect = posix.stat(__file__)
    encoded = jsonpickle.encode(expect)
    actual = jsonpickle.decode(encoded)
    assert expect == actual


def test_repr_using_jsonpickle():
    """Serialize an object that uses jsonpickle in its __repr__ definition"""
    thing = ObjWithJsonPickleRepr()
    thing.child = ObjWithJsonPickleRepr()
    thing.child.parent = thing
    encoded = jsonpickle.encode(thing)
    decoded = jsonpickle.decode(encoded)
    assert id(decoded) == id(decoded.child.parent)


def test_broken_repr_dict_key():
    """Tests that we can pickle dictionaries with keys that have
    broken __repr__ implementations.
    """
    br = BrokenReprThing('test')
    obj = {br: True}
    pickler = jsonpickle.pickler.Pickler()
    flattened = pickler.flatten(obj)
    assert '<BrokenReprThing "test">' in flattened
    assert flattened['<BrokenReprThing "test">']


def test_ordered_dict_python3():
    """Ensure that we preserve dict order on python3"""
    # Python3.6+ preserves dict order.
    obj = {'z': 'Z', 'x': 'X', 'y': 'Y'}
    clone = jsonpickle.decode(jsonpickle.encode(obj))
    expect = ['z', 'x', 'y']
    actual = list(clone.keys())
    assert expect == actual


def test_ordered_dict():
    """Serialize an OrderedDict"""
    d = collections.OrderedDict([('c', 3), ('a', 1), ('b', 2)])
    encoded = jsonpickle.encode(d)
    decoded = jsonpickle.decode(encoded)
    assert d == decoded


def test_ordered_dict_unpicklable():
    """Serialize an OrderedDict with unpicklable=False"""
    d = collections.OrderedDict([('c', 3), ('a', 1), ('b', 2)])
    encoded = jsonpickle.encode(d, unpicklable=False)
    decoded = jsonpickle.decode(encoded)
    assert d == decoded


def test_ordered_dict_reduces():
    """Ensure that OrderedDict is reduce()-able"""
    d = collections.OrderedDict([('c', 3), ('a', 1), ('b', 2)])
    has_reduce, has_reduce_ex = util.has_reduce(d)
    assert util.is_reducible(d)
    assert has_reduce or has_reduce_ex


def test_int_keys_in_object_with_getstate_only():
    """Serialize objects with dict keys that implement __getstate__ only"""
    obj = IntKeysObject()
    encoded = jsonpickle.encode(obj, keys=True)
    decoded = jsonpickle.decode(encoded, keys=True)
    assert obj.data == decoded.data


def test_ordered_dict_int_keys():
    """Serialize dicts with int keys and OrderedDict values"""
    d = {
        1: collections.OrderedDict([(2, -2), (3, -3)]),
        4: collections.OrderedDict([(5, -5), (6, -6)]),
    }
    encoded = jsonpickle.encode(d, keys=True)
    decoded = jsonpickle.decode(encoded, keys=True)
    assert isinstance(decoded[1], collections.OrderedDict)
    assert isinstance(decoded[4], collections.OrderedDict)
    assert -2 == decoded[1][2]
    assert -3 == decoded[1][3]
    assert -5 == decoded[4][5]
    assert -6 == decoded[4][6]
    assert d == decoded


def test_ordered_dict_nested():
    """Serialize nested dicts with OrderedDict values"""
    bottom = collections.OrderedDict([('z', 1), ('a', 2)])
    middle = collections.OrderedDict([('c', bottom)])
    top = collections.OrderedDict([('b', middle)])
    encoded = jsonpickle.encode(top)
    decoded = jsonpickle.decode(encoded)
    assert top == decoded
    # test unpicklable=False
    encoded = jsonpickle.encode(top, unpicklable=False)
    decoded = jsonpickle.decode(encoded)
    assert top == decoded


def test_deque_roundtrip():
    """Make sure we can handle collections.deque"""
    old_deque = collections.deque([0, 1, 2], maxlen=5)
    encoded = jsonpickle.encode(old_deque)
    new_deque = jsonpickle.decode(encoded)
    assert encoded != 'nil'
    assert old_deque[0] == 0
    assert new_deque[0] == 0
    assert old_deque[1] == 1
    assert new_deque[1] == 1
    assert old_deque[2] == 2
    assert new_deque[2] == 2
    assert old_deque.maxlen == 5
    assert new_deque.maxlen == 5


def test_namedtuple_roundtrip():
    """Serialize a NamedTuple"""
    old_nt = NamedTuple(0, 1, 2)
    encoded = jsonpickle.encode(old_nt)
    new_nt = jsonpickle.decode(encoded)
    assert type(old_nt) is type(new_nt)
    assert old_nt is not new_nt
    assert old_nt.a == new_nt.a
    assert old_nt.b == new_nt.b
    assert old_nt.c == new_nt.c
    assert old_nt[0] == new_nt[0]
    assert old_nt[1] == new_nt[1]
    assert old_nt[2] == new_nt[2]


def test_counter_roundtrip():
    counter = collections.Counter({1: 2})
    encoded = jsonpickle.encode(counter)
    decoded = jsonpickle.decode(encoded)
    assert type(decoded) is collections.Counter
    # the integer key becomes a string when keys=False
    assert decoded.get('1') == 2


def test_counter_roundtrip_with_keys():
    counter = collections.Counter({1: 2})
    encoded = jsonpickle.encode(counter, keys=True)
    decoded = jsonpickle.decode(encoded, keys=True)
    assert type(decoded) is collections.Counter
    assert decoded.get(1) == 2


def test_list_with_fd():
    """Serialize a list with an file descriptor"""
    with open(__file__) as fd:
        _test_list_with_fd(fd)
    _test_list_with_fd(fd)  # fd is closed.


def _test_list_with_fd(fd):
    """Serialize a list with an file descriptor"""
    obj = [fd]
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert [None] == newobj


def test_thing_with_fd():
    """Serialize an object with a file descriptor"""
    with open(__file__) as fd:
        _test_thing_with_fd(fd)
    _test_thing_with_fd(fd)  # fd is closed.


def _test_thing_with_fd(fd):
    """Serialize an object with a file descriptor"""
    obj = Thing(fd)
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.name is None


def test_dict_with_fd():
    """Serialize a dict with a file descriptor"""
    with open(__file__) as fd:
        _test_dict_with_fd(fd)
    _test_dict_with_fd(fd)


def _test_dict_with_fd(fd):
    """Serialize a dict with a file descriptor"""
    obj = {'fd': fd}
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj['fd'] is None


def test_thing_with_lamda():
    """Serialize an object with a lambda"""
    obj = Thing(lambda: True)
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert not hasattr(newobj, 'name')


def test_newstyleslots():
    """Serialize an object with new-style slots"""
    obj = ThingWithSlots(True, False)
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.a
    assert not newobj.b


def test_newstyleslots_inherited():
    """Serialize an object with inherited new-style slots"""
    obj = ThingWithInheritedSlots(True, False, None)
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.a
    assert not newobj.b
    assert newobj.c is None


def test_newstyleslots_inherited_deleted_attr():
    """Serialize an object with inherited and deleted new-style slots"""
    obj = ThingWithInheritedSlots(True, False, None)
    del obj.c
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.a
    assert not newobj.b
    assert not hasattr(newobj, 'c')


def test_newstyleslots_with_children():
    """Serialize an object with slots containing objects"""
    obj = ThingWithSlots(Thing('a'), Thing('b'))
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.a.name == 'a'
    assert newobj.b.name == 'b'


def test_newstyleslots_with_children_inherited():
    """Serialize an object with inherited slots containing objects"""
    obj = ThingWithInheritedSlots(Thing('a'), Thing('b'), Thing('c'))
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.a.name == 'a'
    assert newobj.b.name == 'b'
    assert newobj.c.name == 'c'


def test_newstyleslots_iterable():
    """Seriazlie an object with iterable slots"""
    obj = ThingWithIterableSlots('alpha', 'bravo')
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.a == 'alpha'
    assert newobj.b == 'bravo'


def test_newstyleslots_string_slot():
    """Serialize an object with string slots"""
    obj = ThingWithStringSlots('a', 'b')
    jsonstr = jsonpickle.encode(obj)
    newobj = jsonpickle.decode(jsonstr)
    assert newobj.ab == 'ab'


def test_enum34_nested():
    """Serialize enums as nested member variables in an object"""
    ec = EnumClass()
    encoded = jsonpickle.encode(ec)
    decoded = jsonpickle.decode(encoded)
    assert ec.enum_a == decoded.enum_a
    assert ec.enum_b == decoded.enum_b


def test_enum_references():
    """Serialize duplicate enums so that reference IDs are used"""
    a = IntEnumTest.X
    b = IntEnumTest.X
    enums_list = [a, b]
    encoded = jsonpickle.encode(enums_list)
    decoded = jsonpickle.decode(encoded)
    assert enums_list == decoded


def test_enum_unpicklable():
    """Serialize enums when unpicklable=False is specified"""
    obj = Message(MessageTypes.STATUS, MessageCommands.STATUS_ALL)
    encoded = jsonpickle.encode(obj, unpicklable=False)
    decoded = jsonpickle.decode(encoded)
    assert 'message_type' in decoded
    assert 'command' in decoded


def test_enum_int_key_and_value():
    """Serialize Integer enums as dict keys and values"""
    thing = Thing('test')
    value = IntEnumTest.X
    value2 = IntEnumTest.Y
    expect = {
        '__first__': thing,
        'thing': thing,
        value: value,
        value2: value2,
    }
    string = jsonpickle.encode(expect, keys=True)
    actual = jsonpickle.decode(string, keys=True)
    assert 'test' == actual['__first__'].name
    assert value == actual[value]
    assert value2 == actual[value2]

    actual_first = actual['__first__']
    actual_thing = actual['thing']
    assert actual_first is actual_thing


def test_enum_string_key_and_value():
    """Encode enums dict keys and values"""
    thing = Thing('test')
    value = StringEnumTest.A
    value2 = StringEnumTest.B
    expect = {
        value: value,
        '__first__': thing,
        value2: value2,
    }
    string = jsonpickle.encode(expect, keys=True)
    actual = jsonpickle.decode(string, keys=True)
    assert 'test' == actual['__first__'].name
    assert value == actual[value]
    assert value2 == actual[value2]


def test_multiple_string_enums_when_make_refs_is_false():
    """Enums do not create cycles when make_refs=False"""
    # The make_refs=False code path will fallback to repr() when encoding
    # objects that it believes introduce a cycle.  It does this to break
    # out of what would be infinite recursion during traversal.
    # This test ensures that enums do not trigger cycles and are properly
    # encoded under make_refs=False.
    expect = {
        'a': StringEnumTest.A,
        'aa': StringEnumTest.A,
    }
    string = jsonpickle.encode(expect, make_refs=False)
    actual = jsonpickle.decode(string)
    assert expect == actual


# Test classes for ExternalHandlerTestCase
class Mixin:
    def ok(self):
        return True


class UnicodeMixin(str, Mixin):
    def __add__(self, rhs):
        obj = super().__add__(rhs)
        return UnicodeMixin(obj)


class UnicodeMixinHandler(handlers.BaseHandler):
    def flatten(self, obj, data):
        data['value'] = obj
        return data

    def restore(self, obj):
        return UnicodeMixin(obj['value'])


def test_unicode_mixin():
    obj = UnicodeMixin('test')
    assert isinstance(obj, UnicodeMixin)
    assert obj == 'test'

    # Encode into JSON
    handlers.register(UnicodeMixin, UnicodeMixinHandler)
    content = jsonpickle.encode(obj)

    # Resurrect from JSON
    new_obj = jsonpickle.decode(content)
    handlers.unregister(UnicodeMixin)

    new_obj += ' passed'

    assert new_obj == 'test passed'
    assert isinstance(new_obj, UnicodeMixin)
    assert new_obj.ok()


def test_datetime_with_tz_copies_refs():
    """Ensure that we create copies of referenced objects"""
    utc = datetime.timezone.utc
    d0 = datetime.datetime(2020, 5, 5, 5, 5, 5, 5, tzinfo=utc)
    d1 = datetime.datetime(2020, 5, 5, 5, 5, 5, 5, tzinfo=utc)
    obj = [d0, d1]
    encoded = jsonpickle.encode(obj, make_refs=False)
    decoded = jsonpickle.decode(encoded)
    assert len(decoded) == 2
    assert decoded[0] == d0
    assert decoded[1] == d1


def test_with_exclude():
    """Does the _jsonpickle_exclude work"""
    obj = ThingWithExclusion('baz')
    encoded = jsonpickle.encode(obj)
    decoded = jsonpickle.decode(encoded)
    assert decoded.bar == 'baz'
    assert not hasattr(decoded, 'foo')


def test_contained_exclusion():
    """_jsonpickle_exclude should work only on the class it is defined in"""
    obj = ThingWithExcludeSubclass('alpha')
    encoded = jsonpickle.encode(obj)
    decoded = jsonpickle.decode(encoded)
    assert decoded.foo == 'alpha'
    assert not hasattr(decoded.thing, 'foo')



import datetime
import warnings

import pytest

try:
    import numpy as np
    import numpy.testing as npt
    from numpy.testing import assert_equal
except ImportError:
    pytest.skip('numpy is not available', allow_module_level=True)

import jsonpickle
import jsonpickle.ext.numpy
from jsonpickle import handlers


@pytest.fixture(scope='module', autouse=True)
def numpy_extension():
    """Initialize the numpy extension for this test module"""
    jsonpickle.ext.numpy.register_handlers(ndarray_mode='warn')
    yield  # control to the test function.
    jsonpickle.ext.numpy.unregister_handlers()


def roundtrip(obj):
    return jsonpickle.decode(jsonpickle.encode(obj))


def test_dtype_roundtrip():
    dtypes = [
        np.int_,
        np.int32,
        np.float64,
        np.complex128,
        np.str_,
        np.object_,
        np.dtype(np.void),
        np.dtype(np.int32),
        np.dtype(np.float32),
        np.dtype('f4,i4,f2,i1'),
        np.dtype(('f4', 'i4'), ('f2', 'i1')),
        np.dtype('i4', align=True),
        np.dtype('M8[7D]'),
        np.dtype(
            {
                'names': ['f0', 'f1', 'f2'],
                'formats': ['<u4', '<u2', '<u2'],
                'offsets': [0, 0, 2],
            },
            align=True,
        ),
        np.dtype([('f0', 'i4'), ('f2', 'i1')]),
        np.dtype(
            [
                (
                    'top',
                    [
                        ('tiles', ('>f4', (64, 64)), (1,)),
                        ('rtile', '>f4', (64, 36)),
                    ],
                    (3,),
                ),
                (
                    'bottom',
                    [
                        ('bleft', ('>f4', (8, 64)), (1,)),
                        ('bright', '>f4', (8, 36)),
                    ],
                ),
            ]
        ),
    ]

    for dtype in dtypes:
        encoded = jsonpickle.encode(dtype)
        decoded = jsonpickle.decode(encoded)
        assert dtype == decoded


def test_generic_roundtrip():
    values = [
        np.int_(1),
        np.int32(-2),
        np.float64(2.5),
        np.nan,
        -np.inf,
        np.inf,
        np.datetime64('2014-01-01'),
        np.str_('foo'),
        np.object_({'a': 'b'}),
        np.complex128(1 - 2j),
    ]
    for value in values:
        decoded = roundtrip(value)
        assert_equal(decoded, value)
        assert isinstance(decoded, type(value))


def test_ndarray_roundtrip():
    arrays = [
        np.random.random((10, 20)),
        np.array([[True, False, True]]),
        np.array(['foo', 'bar']),
        np.array([b'baz']),
        np.array(['2010', 'NaT', '2030']).astype('M'),
        np.rec.array(b'abcdefg' * 100, formats='i2,a3,i4', shape=3),
        np.rec.array(
            [
                (1, 11, 'a'),
                (2, 22, 'b'),
                (3, 33, 'c'),
                (4, 44, 'd'),
                (5, 55, 'ex'),
                (6, 66, 'f'),
                (7, 77, 'g'),
            ],
            formats='u1,f4,a1',
        ),
        np.array(['1960-03-12', datetime.date(1960, 3, 12)], dtype='M8[D]'),
        np.array([0, 1, -1, np.inf, -np.inf, np.nan], dtype='f2'),
    ]

    arrays.extend(
        [
            np.rec.array(
                [('NGC1001', 11), ('NGC1002', 1.0), ('NGC1003', 1.0)],
                dtype=[('target', 'S20'), ('V_mag', 'f4')],
            )
        ]
    )
    for array in arrays:
        decoded = roundtrip(array)
        assert_equal(decoded, array)
        assert decoded.dtype == array.dtype


def test_shapes_containing_zeroes():
    """Test shapes which cannot be represented as nested lists"""
    expect = np.eye(3)[3:]
    actual = roundtrip(expect)
    npt.assert_array_equal(expect, actual)


def test_accuracy():
    """Test the accuracy of the string representation"""
    expect = np.random.randn(3, 3)
    actual = roundtrip(expect)
    npt.assert_array_equal(expect, actual)


def test_b64():
    """Test the binary encoding"""
    # Array of substantial size is stored as b64.
    expect = np.random.rand(10, 10)
    actual = roundtrip(expect)
    npt.assert_array_equal(expect, actual)


def test_views():
    """Test views under serialization"""
    rng = np.arange(20)  # a range of an array
    view = rng[10:]  # a view referencing a portion of an array
    data = [rng, view]

    actual = roundtrip(data)
    actual[0][15] = -1
    assert actual[1][5] == -1


def test_strides():
    """Test non-standard strides and offsets"""
    arr = np.eye(3)
    view = arr[1:, 1:]
    assert view.base is arr

    data = [arr, view]
    actual = roundtrip(data)

    # test that the deserialized arrays indeed view the same memory
    new_arr, new_view = actual
    new_arr[1, 2] = -1
    assert new_view[0, 1] == -1
    assert new_view.base is new_arr


def test_weird_arrays():
    """Test references to arrays that do not effectively own their memory"""
    a = np.arange(9)
    b = a[5:]
    a.strides = 1

    # this is kinda fishy; a has overlapping memory, _a does not
    warn_count = 1
    with warnings.catch_warnings(record=True) as w:
        _a = roundtrip(a)
        assert len(w) == warn_count
        npt.assert_array_equal(a, _a)

    # this also requires a deepcopy to work
    with warnings.catch_warnings(record=True) as w:
        _a, _b = roundtrip([a, b])
        assert len(w) == warn_count
        npt.assert_array_equal(a, _a)
        npt.assert_array_equal(b, _b)


def test_transpose():
    """test handling of non-c-contiguous memory layout"""
    # simple case; view a c-contiguous array
    a = np.arange(9).reshape(3, 3)
    b = a[1:, 1:]
    assert b.base is a.base
    _a, _b = roundtrip([a, b])
    assert _b.base is _a.base
    npt.assert_array_equal(a, _a)
    npt.assert_array_equal(b, _b)

    # a and b both view the same contiguous array
    a = np.arange(9).reshape(3, 3).T
    b = a[1:, 1:]
    assert b.base is a.base
    _a, _b = roundtrip([a, b])
    assert _b.base is _a.base
    npt.assert_array_equal(a, _a)
    npt.assert_array_equal(b, _b)

    # view an f-contiguous array
    a = a.copy()
    a.strides = a.strides[::-1]
    b = a[1:, 1:]
    assert b.base is a
    _a, _b = roundtrip([a, b])
    assert _b.base is _a
    npt.assert_array_equal(a, _a)
    npt.assert_array_equal(b, _b)

    # now a.data.contiguous is False; we have to make a deepcopy to make
    # this work note that this is a pretty contrived example though!
    a = np.arange(8).reshape(2, 2, 2).copy()
    a.strides = a.strides[0], a.strides[2], a.strides[1]
    b = a[1:, 1:]
    assert b.base is a

    warn_count = 1
    with warnings.catch_warnings(record=True) as w:
        _a, _b = roundtrip([a, b])
        assert len(w) == warn_count
        npt.assert_array_equal(a, _a)
        npt.assert_array_equal(b, _b)


def test_fortran_base():
    """Test a base array in fortran order"""
    a = np.asfortranarray(np.arange(100).reshape((10, 10)))
    _a = roundtrip(a)
    npt.assert_array_equal(a, _a)


def test_buffer():
    """test behavior with memoryviews which are not ndarrays"""
    bstring = b'abcdefgh'
    a = np.frombuffer(bstring, dtype=np.byte)
    warn_count = 1
    with warnings.catch_warnings(record=True) as w:
        _a = roundtrip(a)
        npt.assert_array_equal(a, _a)
        assert len(w) == warn_count


def test_as_strided():
    """Test the result of as_strided()

    as_strided() returns an object that implements the array interface but
    is not an ndarray.

    """
    warn_count = 1
    a = np.arange(10)
    b = np.lib.stride_tricks.as_strided(a, shape=(5,), strides=(a.dtype.itemsize * 2,))
    data = [a, b]

    with warnings.catch_warnings(record=True) as w:
        # as_strided returns a DummyArray object, which we can not
        # currently serialize correctly FIXME: would be neat to add
        # support for all objects implementing the __array_interface__
        _data = roundtrip(data)
        assert len(w) == warn_count

    # as we were warned, deserialized result is no longer a view
    _data[0][0] = -1
    assert _data[1][0] == 0


def test_immutable():
    """test that immutability flag is copied correctly"""
    a = np.arange(10)
    a.flags.writeable = False
    _a = roundtrip(a)
    with pytest.raises(ValueError):
        _a[0] = 0


def test_byteorder():
    """Test the byteorder for text and binary encodings"""
    # small arr is stored as text
    a = np.arange(10)
    av = a.view(a.dtype.newbyteorder())
    b = a[:]
    bv = b.view(b.dtype.newbyteorder())
    _av, _bv = roundtrip([av, bv])
    npt.assert_array_equal(av, _av)
    npt.assert_array_equal(bv, _bv)

    # bigger arr is stored as binary
    a = np.arange(100)
    av = a.view(a.dtype.newbyteorder())
    b = a[:]
    bv = b.view(b.dtype.newbyteorder())
    _av, _bv = roundtrip([av, bv])
    npt.assert_array_equal(av, _av)
    npt.assert_array_equal(bv, _bv)


def test_zero_dimensional_array():
    expect = np.array(float(0.0), dtype="float64")
    actual = jsonpickle.decode(jsonpickle.encode(expect))
    npt.assert_array_equal(expect, actual)


def test_nested_data_list_of_dict_with_list_keys():
    """Ensure we can handle numpy arrays within a nested structure"""
    expect = [{'key': [np.array(0)]}]
    actual = roundtrip(expect)
    npt.assert_array_equal(expect[0]['key'][0], actual[0]['key'][0])

    expect = [{'key': [np.array([1.0])]}]
    actual = roundtrip(expect)
    npt.assert_array_equal(expect[0]['key'][0], actual[0]['key'][0])


def test_size_threshold_None():
    handler = jsonpickle.ext.numpy.NumpyNDArrayHandlerView(size_threshold=None)
    handlers.registry.unregister(np.ndarray)
    handlers.registry.register(np.ndarray, handler, base=True)
    expect = np.array([0, 1])
    actual = roundtrip(expect)
    npt.assert_array_equal(expect, actual)


def test_ndarray_dtype_object():
    a = np.array(['F' + str(i) for i in range(30)], dtype=object)
    buf = jsonpickle.encode(a)
    # This is critical for reproducing the numpy segfault issue when
    # restoring ndarray of dtype object.
    del a
    expect = np.array(['F' + str(i) for i in range(30)], dtype=object)
    actual = jsonpickle.decode(buf)
    npt.assert_array_equal(expect, actual)


def test_np_random():
    """Ensure random.random() arrays can be serialized"""
    obj = np.random.random(100)
    encoded = jsonpickle.encode(obj)
    clone = jsonpickle.decode(encoded)
    assert 100 == len(clone)
    for idx, (expect, actual) in enumerate(zip(obj, clone)):
        assert expect == actual


def test_np_poly1d():
    # issue 391, test poly1d roundtrip
    obj = np.poly1d([1, 2, 3])
    assert obj == jsonpickle.decode(jsonpickle.encode(obj))



# Copyright (C) 2013 Jason R. Coombs <jaraco@jaraco.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

import unittest

import pytest

import jsonpickle
import jsonpickle.handlers


class CustomObject:
    "A class to be serialized by a custom handler"

    def __init__(self, name=None, creator=None):
        self.name = name
        self.creator = creator

    def __eq__(self, other):
        return self.name == other.name


class CustomA(CustomObject):
    pass


class CustomB(CustomA):
    pass


class NullHandler(jsonpickle.handlers.BaseHandler):
    def flatten(self, obj, data):
        data['name'] = obj.name
        return data

    def restore(self, obj):
        return CustomObject(obj['name'], creator=type(self))


class DecoratedBase(CustomObject):
    pass


class DecoratedChild(DecoratedBase):
    pass


@jsonpickle.handlers.register(DecoratedBase, base=True)
class DecoratedHandler(NullHandler):
    pass


class SithHandler(jsonpickle.handlers.BaseHandler):
    """(evil) serialization handler that rewrites the entire object"""

    def flatten(self, obj, data):
        data['name'] = obj.name
        data['py/object'] = 'sith.lord'
        return data


class HandlerTestCase(unittest.TestCase):
    def setUp(self):
        jsonpickle.handlers.register(CustomObject, NullHandler)

    def tearDown(self):
        jsonpickle.handlers.unregister(CustomObject)

    def roundtrip(self, ob):
        encoded = jsonpickle.encode(ob)
        decoded = jsonpickle.decode(encoded)
        assert decoded == ob
        return decoded

    def test_custom_handler(self):
        """Ensure that the custom handler is indeed used"""
        expect = CustomObject('jarjar')
        encoded = jsonpickle.encode(expect)
        actual = jsonpickle.decode(encoded)
        assert expect.name == actual.name
        assert expect.creator is None
        assert actual.creator is NullHandler

    def test_references(self):
        """
        Ensure objects handled by a custom handler are properly dereferenced.
        """
        ob = CustomObject()
        # create a dictionary which contains several references to ob
        subject = dict(a=ob, b=ob, c=ob)
        # ensure the subject can be roundtripped
        new_subject = self.roundtrip(subject)
        assert new_subject['a'] == new_subject['b']
        assert new_subject['b'] == new_subject['c']
        assert new_subject['a'] is new_subject['b']
        assert new_subject['b'] is new_subject['c']

    def test_invalid_class(self):
        with pytest.raises(TypeError):
            jsonpickle.handlers.register('foo', NullHandler)

    def test_base_handler(self):
        a = CustomA('a')
        assert a.creator is None
        assert jsonpickle.decode(jsonpickle.encode(a)).creator is None

        b = CustomB('b')
        assert b.creator is None
        assert jsonpickle.decode(jsonpickle.encode(b)).creator is None

        OtherHandler = type('OtherHandler', (NullHandler,), {})
        jsonpickle.handlers.register(CustomA, OtherHandler, base=True)
        assert self.roundtrip(a).creator is OtherHandler
        assert self.roundtrip(b).creator is OtherHandler

        SpecializedHandler = type('SpecializedHandler', (NullHandler,), {})
        jsonpickle.handlers.register(CustomB, SpecializedHandler)
        assert self.roundtrip(a).creator is OtherHandler
        assert self.roundtrip(b).creator is SpecializedHandler

    def test_decorated_register(self):
        db = DecoratedBase('db')
        dc = DecoratedChild('dc')
        assert self.roundtrip(db).creator is DecoratedHandler
        assert self.roundtrip(dc).creator is DecoratedHandler

    def test_custom_handler_can_rewrite_everything(self):
        """Test the low-level pickling structures"""
        jsonpickle.handlers.unregister(CustomObject)
        jsonpickle.handlers.register(CustomObject, SithHandler)
        obj = CustomObject('jarjar')  # secret sith lord jarjar
        pickler = jsonpickle.pickler.Pickler()
        data = pickler.flatten(obj)
        assert isinstance(data, dict)
        assert len(data) == 2
        assert data['name'] == 'jarjar'
        assert data['py/object'] == 'sith.lord'



#
# Copyright (C) 2013 Jason R. Coombs <jaraco@jaraco.com>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution.

import datetime
import sys
import time
import unittest

import jsonpickle
from jsonpickle import tags


class ObjWithDate:
    def __init__(self):
        ts = datetime.datetime.now()
        self.data = dict(a='a', ts=ts)
        self.data_ref = dict(b='b', ts=ts)


# UTC implementation from Python 2.7 docs
class UTC(datetime.tzinfo):
    """UTC"""

    def utcoffset(self, dt):
        return datetime.timedelta()

    def tzname(self, dt):
        return 'UTC'

    def dst(self, dt):
        return datetime.timedelta()


utc = UTC()


class TimestampedVariable:
    def __init__(self, value=None):
        self._value = value
        self._dt_read = datetime.datetime.now(tz=datetime.timezone.utc)
        self._dt_write = self._dt_read

    def get(self, default_value=None):
        if self._dt_read is None and self._dt_write is None:
            value = default_value
            self._value = value
            self._dt_write = datetime.datetime.now(tz=datetime.timezone.utc)
        else:
            value = self._value
        self._dt_read = datetime.datetime.now(tz=datetime.timezone.utc)
        return value

    def set(self, new_value):
        self._dt_write = datetime.datetime.now(tz=datetime.timezone.utc)
        self._value = new_value

    def erasable(self, td=datetime.timedelta(seconds=1)):
        dt_now = datetime.datetime.now(tz=datetime.timezone.utc)
        td_read = dt_now - self._dt_read
        td_write = dt_now - self._dt_write
        return td_read > td and td_write > td

    def __repr__(self):
        dt_now = datetime.datetime.now(tz=datetime.timezone.utc)
        td_read = dt_now - self._dt_read
        td_write = dt_now - self._dt_write
        s = '<TimestampedVariable>\n'
        s += ' value: ' + str(self._value) + '\n'
        s += ' dt_read: ' + str(self._dt_read) + ' (%s ago)' % td_read + '\n'
        s += ' dt_write: ' + str(self._dt_write) + ' (%s ago)' % td_write + '\n'
        return s


class PersistantVariables:
    def __init__(self):
        self._data = {}

    def __getitem__(self, key):
        return self._data.setdefault(key, TimestampedVariable(None))

    def __setitem__(self, key, value):
        return self._data.setdefault(key, TimestampedVariable(value))

    def __repr__(self):
        return str(self._data)


def test_object_with_inner_datetime_refs():
    pvars = PersistantVariables()
    pvars['z'] = 1
    pvars['z2'] = 2
    pickled = jsonpickle.encode(pvars)
    obj = jsonpickle.decode(pickled)
    # ensure the references are valid
    assert obj['z']._dt_read is obj['z']._dt_write
    assert obj['z2']._dt_read is obj['z2']._dt_write
    # ensure the values are valid
    assert obj['z'].get() == 1
    assert obj['z2'].get() == 2
    # ensure get() updates _dt_read
    assert obj['z']._dt_read is not obj['z']._dt_write
    assert obj['z2']._dt_read is not obj['z2']._dt_write


def _roundtrip(obj):
    """Roundtrip encode and decode an object and assert equality"""
    pickled = jsonpickle.encode(obj)
    unpickled = jsonpickle.decode(pickled)
    assert obj == unpickled


def test_datetime():
    """Roundtrip datetime objects"""
    _roundtrip(datetime.datetime.now())


def test_date():
    """Roundtrip date objects"""
    _roundtrip(datetime.datetime.today())


def test_time():
    """Roundtrip time objects"""
    _roundtrip(datetime.datetime.now().time())


def test_timedelta():
    """Roundtrip timedelta objects"""
    _roundtrip(datetime.timedelta(days=3))


def test_utc():
    """Roundtrip datetime objectcs with UTC timezone info"""
    _roundtrip(datetime.datetime.now(tz=datetime.timezone.utc).replace(tzinfo=utc))


def test_unpickleable():
    """Date objects are human-readable strings when unpicklable is False"""
    obj = datetime.datetime.now()
    pickler = jsonpickle.pickler.Pickler(unpicklable=False)
    flattened = pickler.flatten(obj)
    assert obj.isoformat() == flattened


def test_object_with_datetime():
    test_obj = ObjWithDate()
    json = jsonpickle.encode(test_obj)
    test_obj_decoded = jsonpickle.decode(json)
    assert test_obj_decoded.data['ts'] == test_obj_decoded.data_ref['ts']


@unittest.skipIf(sys.version_info < (3, 9), 'only tested for python >= 3.9')
def test_datetime_with_zoneinfo():
    """Roundtrip datetime objects with ZoneInfo tzinfo"""
    from zoneinfo import ZoneInfo

    now = datetime.datetime.now()
    SaoPaulo = ZoneInfo('America/Sao_Paulo')
    NewYork = ZoneInfo('America/New_York')
    now_sp = now.replace(tzinfo=SaoPaulo)
    now_us = now.replace(tzinfo=NewYork)
    _roundtrip(now_sp)
    _roundtrip(now_us)


def test_struct_time():
    expect = time.struct_time([1, 2, 3, 4, 5, 6, 7, 8, 9])
    json = jsonpickle.encode(expect)
    actual = jsonpickle.decode(json)
    assert type(actual) is time.struct_time
    assert expect == actual


def test_struct_time_chars():
    pickler = jsonpickle.pickler.Pickler()
    unpickler = jsonpickle.unpickler.Unpickler()
    expect = time.struct_time('123456789')
    flattened = pickler.flatten(expect)
    actual = unpickler.restore(flattened)
    assert expect == actual


def test_datetime_structure():
    pickler = jsonpickle.pickler.Pickler()
    unpickler = jsonpickle.unpickler.Unpickler()
    obj = datetime.datetime.now()
    flattened = pickler.flatten(obj)
    assert tags.OBJECT in flattened
    assert '__reduce__' in flattened
    inflated = unpickler.restore(flattened)
    assert obj == inflated


def test_datetime_inside_int_keys_defaults():
    t = datetime.time(hour=10)
    s = jsonpickle.encode({1: t, 2: t})
    d = jsonpickle.decode(s)
    assert d['1'] == d['2']
    assert d['1'] is d['2']
    assert isinstance(d['1'], datetime.time)


def test_datetime_inside_int_keys_with_keys_enabled():
    t = datetime.time(hour=10)
    s = jsonpickle.encode({1: t, 2: t}, keys=True)
    d = jsonpickle.decode(s, keys=True)
    assert d[1] == d[2]
    assert d[1] is d[2]
    assert isinstance(d[1], datetime.time)


def test_datetime_repr_not_unpicklable():
    obj = datetime.datetime.now()
    pickler = jsonpickle.pickler.Pickler(unpicklable=False)
    flattened = pickler.flatten(obj)
    assert tags.REPR not in flattened
    assert tags.MODULE not in flattened
    assert tags.OBJECT not in flattened
    assert obj.isoformat() == flattened


def test_datetime_dict_keys_defaults():
    """Test that we handle datetime objects as keys."""
    datetime_dict = {datetime.datetime(2008, 12, 31): True}
    pickled = jsonpickle.encode(datetime_dict)
    expect = {'datetime.datetime(2008, 12, 31, 0, 0)': True}
    actual = jsonpickle.decode(pickled)
    assert expect == actual


def test_datetime_dict_keys_with_keys_enabled():
    """Test that we handle datetime objects as keys."""
    datetime_dict = {datetime.datetime(2008, 12, 31): True}
    pickled = jsonpickle.encode(datetime_dict, keys=True)
    expect = datetime_dict
    actual = jsonpickle.decode(pickled, keys=True)
    assert expect == actual



import jsonpickle


class Node:
    def __init__(self, name):
        self._name = name
        self._children = []
        self._parent = None

    def add_child(self, child, index=-1):
        if index == -1:
            index = len(self._children)
        self._children.insert(index, child)
        child._parent = self


class Document(Node):
    def __init__(self, name):
        Node.__init__(self, name)

    def __repr__(self):
        return str(self)

    def __str__(self):
        ret_str = 'Document "%s"\n' % self._name
        for c in self._children:
            ret_str += repr(c)
        return ret_str


class Question(Node):
    def __init__(self, name):
        Node.__init__(self, name)

    def __str__(self):
        return f'Question "{self._name}", parent: "{self._parent._name}"\n'

    def __repr__(self):
        return self.__str__()


class Section(Node):
    def __init__(self, name):
        Node.__init__(self, name)

    def __str__(self):
        ret_str = f'Section "{self._name}", parent: "{self._parent._name}"\n'
        for c in self._children:
            ret_str += repr(c)
        return ret_str

    def __repr__(self):
        return self.__str__()


def test_cyclical():
    """Test that we can pickle cyclical data structure

    This test is ensures that we can reference objects which
    first appear within a list (in other words, not a top-level
    object or attribute).  Later children will reference that
    object through its "_parent" field.

    This makes sure that we handle this case correctly.

    """
    document = Document('My Document')
    section1 = Section('Section 1')
    section2 = Section('Section 2')
    question1 = Question('Question 1')
    question2 = Question('Question 2')
    question3 = Question('Question 3')
    question4 = Question('Question 4')
    document.add_child(section1)
    document.add_child(section2)
    section1.add_child(question1)
    section1.add_child(question2)
    section2.add_child(question3)
    section2.add_child(question4)
    pickled = jsonpickle.encode(document)
    unpickled = jsonpickle.decode(pickled)
    assert str(document) == str(unpickled)



from utilities import ensure_no_files_overwritten

import jsonpickle

ensure_no_files_overwritten(
    expected_contents='{"py/object": "__main__.Example", "data": {"BAR": 1, "foo": 0}}'
)


class Example:
    def __init__(self):
        self.data = {"foo": 0, "BAR": 1}

    def get_foo(self):
        return self.data["foo"]

    def __eq__(self, other):
        # ensure that jsonpickle preserves important stuff like data across encode/decoding
        return self.data == other.data and self.get_foo() == other.get_foo()


# instantiate the class
ex = Example()
# encode the class. this returns a string, like json.dumps
encoded_instance = jsonpickle.encode(ex)
assert (
    encoded_instance
    == '{"py/object": "__main__.Example", "data": {"BAR": 1, "foo": 0}}'
)
print(
    f"jsonpickle successfully encoded the instance of the Example class! It looks like: {encoded_instance}"
)


with open("example.json", "w+") as f:
    f.write(encoded_instance)
    print(
        "jsonpickle successfully wrote the instance of the Example class to example.json!"
    )

with open("example.json", "r+") as f:
    written_instance = f.read()
    # decode the file into a copy of the original object
    decoded_instance = jsonpickle.decode(written_instance)
    print("jsonpickle successfully decoded the instance of the Example class!")

# this should be true, if it isn't then you should file a bug report
assert decoded_instance == ex
print(
    "jsonpickle successfully decoded the instance of the Example class, and it matches the original instance!"
)



import datetime

import pytest

try:
    import numpy as np
    import pandas as pd
    from pandas.testing import (
        assert_frame_equal,
        assert_index_equal,
        assert_series_equal,
    )
except ImportError:
    pytest.skip('numpy is not available', allow_module_level=True)


import jsonpickle
import jsonpickle.ext.pandas


@pytest.fixture(scope='module', autouse=True)
def pandas_extension():
    """Initialize the numpy extension for this test module"""
    jsonpickle.ext.pandas.register_handlers()
    yield  # control to the test function.
    jsonpickle.ext.pandas.unregister_handlers()


def roundtrip(obj, **kwargs):
    return jsonpickle.decode(jsonpickle.encode(obj, **kwargs))


def test_series_roundtrip():
    ser = pd.Series(
        {
            'an_int': np.int_(1),
            'a_float': np.float64(2.5),
            'a_nan': np.nan,
            'a_minus_inf': -np.inf,
            'an_inf': np.inf,
            'a_str': np.str_('foo'),
            'date': np.datetime64('2014-01-01'),
            'complex': np.complex128(1 - 2j),
            # TODO: the following dtypes are not currently supported.
            # 'object': np.object_({'a': 'b'}),
        }
    )
    decoded_ser = roundtrip(ser)
    assert_series_equal(decoded_ser, ser)


def test_dataframe_roundtrip():
    df = pd.DataFrame(
        {
            'an_int': np.int_([1, 2, 3]),
            'a_float': np.float64([2.5, 3.5, 4.5]),
            'a_nan': np.array([np.nan] * 3),
            'a_minus_inf': np.array([-np.inf] * 3),
            'an_inf': np.array([np.inf] * 3),
            'a_str': np.str_('foo'),
            'date': np.array([np.datetime64('2014-01-01')] * 3, dtype="datetime64[s]"),
            'date_ns': np.array(
                [np.datetime64('2014-01-01')] * 3, dtype="datetime64[ns]"
            ),
            'timedelta': np.array([np.timedelta64(1, "Y")] * 3, dtype="timedelta64[Y]"),
            'complex': np.complex128([1 - 2j, 2 - 1.2j, 3 - 1.3j]),
            # TODO: the following dtypes are not currently supported.
            # 'object': np.object_([{'a': 'b'}]*3),
        }
    )
    decoded_df = roundtrip(df)
    assert_frame_equal(decoded_df, df)


def test_dataframe_nested_containers():
    # adapted from #407
    a = pd.DataFrame(
        {"date": ["20220921abc", 20220921.5], "date2": [20220921, "20220921"]}
    )
    b = roundtrip(a, keys=True)
    assert a['date'][0] == b['date'][0]


def test_dataframe_mixed_dtypes():
    # adapted from #358
    df = pd.DataFrame({'col1': [[1, 2, 3], ['a', 'b', 'c']]})
    assert isinstance((df['col1'].iloc[0]), list)
    result = roundtrip(df)
    assert isinstance((result['col1'].iloc[0]), list)

    # adapted from #457
    # Create simple data frame of mixed data types
    df1 = pd.DataFrame(([1, 2], [4, 'foo']))
    df2 = roundtrip(df1)
    assert (df1[0] == df2[0]).all()


def test_multindex_dataframe_roundtrip():
    df = pd.DataFrame(
        {
            'idx_lvl0': ['a', 'b', 'c'],
            'idx_lvl1': np.int_([1, 1, 2]),
            'an_int': np.int_([1, 2, 3]),
            'a_float': np.float64([2.5, 3.5, 4.5]),
            'a_nan': np.array([np.nan] * 3),
            'a_minus_inf': np.array([-np.inf] * 3),
            'an_inf': np.array([np.inf] * 3),
            'a_str': np.str_('foo'),
        }
    )
    df = df.set_index(['idx_lvl0', 'idx_lvl1'])

    decoded_df = roundtrip(df)
    assert_frame_equal(decoded_df, df)


def test_dataframe_with_interval_index_roundtrip():
    df = pd.DataFrame(
        {'a': [1, 2], 'b': [3, 4]}, index=pd.IntervalIndex.from_breaks([1, 2, 4])
    )

    decoded_df = roundtrip(df)
    assert_frame_equal(decoded_df, df)


def test_index_roundtrip():
    idx = pd.Index(range(5, 10))
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_datetime_index_roundtrip():
    idx = pd.date_range(start='2019-01-01', end='2019-02-01', freq='D')
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_ragged_datetime_index_roundtrip():
    idx = pd.DatetimeIndex(['2019-01-01', '2019-01-02', '2019-01-05'])
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_timedelta_index_roundtrip():
    idx = pd.timedelta_range(start='1 day', periods=4, closed='right')
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_period_index_roundtrip():
    idx = pd.period_range(start='2017-01-01', end='2018-01-01', freq='M')
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_int64_index_roundtrip():
    idx = pd.Index([-1, 0, 3, 4], dtype=np.int64)
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_uint64_index_roundtrip():
    idx = pd.Index([0, 3, 4], dtype=np.uint64)
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_float64_index_roundtrip():
    idx = pd.Index([0.1, 3.7, 4.2], dtype=np.float64)
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_interval_index_roundtrip():
    idx = pd.IntervalIndex.from_breaks(range(5))
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_datetime_interval_index_roundtrip():
    idx = pd.IntervalIndex.from_breaks(pd.date_range('2019-01-01', '2019-01-10'))
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_multi_index_roundtrip():
    idx = pd.MultiIndex.from_product(((1, 2, 3), ('a', 'b')))
    decoded_idx = roundtrip(idx)
    assert_index_equal(decoded_idx, idx)


def test_timestamp_roundtrip():
    obj = pd.Timestamp('2019-01-01')
    decoded_obj = roundtrip(obj)
    assert decoded_obj == obj


def test_period_roundtrip():
    obj = pd.Timestamp('2019-01-01')
    decoded_obj = roundtrip(obj)
    assert decoded_obj == obj


def test_interval_roundtrip():
    obj = pd.Interval(2, 4, closed='left')
    decoded_obj = roundtrip(obj)
    assert decoded_obj == obj


def test_b64():
    """Test the binary encoding"""
    # array of substantial size is stored as b64
    a = np.random.rand(20, 10)
    index = ['Row' + str(i) for i in range(1, a.shape[0] + 1)]
    columns = ['Col' + str(i) for i in range(1, a.shape[1] + 1)]
    df = pd.DataFrame(a, index=index, columns=columns)
    decoded_df = roundtrip(df)
    assert_frame_equal(decoded_df, df)


def test_series_list_index():
    """Test pandas using series with a list index"""
    expect = pd.Series(0, index=[1, 2, 3])
    actual = roundtrip(expect)

    assert expect.values[0] == actual.values[0]
    assert 0 == actual.values[0]

    assert expect.index[0] == actual.index[0]
    assert expect.index[1] == actual.index[1]
    assert expect.index[2] == actual.index[2]


def test_series_multi_index():
    """Test pandas using series with a multi-index"""
    expect = pd.Series(0, index=[[1], [2], [3]])
    actual = roundtrip(expect)

    assert expect.values[0] == actual.values[0]
    assert 0 == actual.values[0]

    assert expect.index[0] == actual.index[0]
    assert expect.index[0][0] == actual.index[0][0]
    assert expect.index[0][1] == actual.index[0][1]
    assert expect.index[0][2] == actual.index[0][2]


def test_series_multi_index_strings():
    """Test multi-index with strings"""
    lets = ['A', 'B', 'C']
    nums = ['1', '2', '3']
    midx = pd.MultiIndex.from_product([lets, nums])
    expect = pd.Series(0, index=midx)
    actual = roundtrip(expect)

    assert expect.values[0] == actual.values[0]
    assert 0 == actual.values[0]

    assert expect.index[0] == actual.index[0]
    assert expect.index[1] == actual.index[1]
    assert expect.index[2] == actual.index[2]
    assert expect.index[3] == actual.index[3]
    assert expect.index[4] == actual.index[4]
    assert expect.index[5] == actual.index[5]
    assert expect.index[6] == actual.index[6]
    assert expect.index[7] == actual.index[7]
    assert expect.index[8] == actual.index[8]

    assert ('A', '1') == actual.index[0]
    assert ('A', '2') == actual.index[1]
    assert ('A', '3') == actual.index[2]
    assert ('B', '1') == actual.index[3]
    assert ('B', '2') == actual.index[4]
    assert ('B', '3') == actual.index[5]
    assert ('C', '1') == actual.index[6]
    assert ('C', '2') == actual.index[7]
    assert ('C', '3') == actual.index[8]


def test_dataframe_with_timedelta64_dtype():
    data_frame = pd.DataFrame(
        {
            'Start': [
                '2020/12/14 00:00:01',
                '2020/12/14 00:00:04',
                '2020/12/14 00:00:06',
            ],
            'End': [
                '2020/12/14 00:00:04',
                '2020/12/14 00:00:06',
                '2020/12/14 00:00:09',
            ],
        }
    )
    data_frame['Start'] = pd.to_datetime(data_frame['Start'])
    data_frame['End'] = pd.to_datetime(data_frame['End'])
    data_frame['Duration'] = data_frame['End'] - data_frame['Start']

    encoded = jsonpickle.encode(data_frame)
    actual = jsonpickle.decode(encoded)

    assert isinstance(actual, pd.DataFrame)
    assert data_frame['Start'][0] == actual['Start'][0]
    assert data_frame['Start'][1] == actual['Start'][1]
    assert data_frame['Start'][2] == actual['Start'][2]
    assert data_frame['End'][0] == actual['End'][0]
    assert data_frame['End'][1] == actual['End'][1]
    assert data_frame['End'][2] == actual['End'][2]
    assert isinstance(actual['Duration'][0], datetime.timedelta)
    assert isinstance(actual['Duration'][1], datetime.timedelta)
    assert isinstance(actual['Duration'][2], datetime.timedelta)
    assert data_frame['Duration'][0] == actual['Duration'][0]
    assert data_frame['Duration'][1] == actual['Duration'][1]
    assert data_frame['Duration'][2] == actual['Duration'][2]


def test_multilevel_columns():
    iterables = [['inj', 'prod'], ['hourly', 'cumulative']]
    names = ['first', 'second']
    # transform it to tuples
    columns = pd.MultiIndex.from_product(iterables, names=names)
    # build a multi-index from it
    data_frame = pd.DataFrame(
        np.random.randn(3, 4), index=['A', 'B', 'C'], columns=columns
    )
    encoded = jsonpickle.encode(data_frame)
    cloned_data_frame = jsonpickle.decode(encoded)
    assert isinstance(cloned_data_frame, pd.DataFrame)
    assert data_frame.columns.names == cloned_data_frame.columns.names
    assert_frame_equal(data_frame, cloned_data_frame)
    assert names == cloned_data_frame.columns.names



"""Test miscellaneous objects from the standard library"""

import uuid

import jsonpickle


def test_random_uuid():
    u = uuid.uuid4()
    encoded = jsonpickle.encode(u)
    decoded = jsonpickle.decode(encoded)

    expect = u.hex
    actual = decoded.hex
    assert expect == actual


def test_known_uuid():
    expect = '28b56adbd18f44e2a5556bba2f23e6f6'
    exemplar = uuid.UUID(expect)
    encoded = jsonpickle.encode(exemplar)
    decoded = jsonpickle.decode(encoded)

    actual = decoded.hex
    assert expect == actual


def test_bytestream():
    expect = (
        b'\x89HDF\r\n\x1a\n\x00\x00\x00\x00\x00\x08\x08\x00'
        b'\x04\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        b'\x00\x00\x00\x00\xff\xff\xff\xff\xff\xff\xff\xffh'
        b'\x848\x00\x00\x00\x00\x00\xff\xff\xff\xff\xff\xff'
        b'\xff\xff\x00\x00\x00\x00\x00\x00\x00\x00`\x00\x00'
        b'\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00'
        b'\x00\x88\x00\x00\x00\x00\x00\x00\x00\xa8\x02\x00'
        b'\x00\x00\x00\x00\x00\x01\x00\x01\x00'
    )
    encoded = jsonpickle.encode(expect)
    actual = jsonpickle.decode(encoded)
    assert expect == actual
